"""

Base class for undirected graphs.

The Graph class allows any hashable object as a node
and can associate any object with an undirected edge.
Self-loops are allowed but multiple edges are not (see MultiGraph)
For directed graphs see DiGraph and XDiGraph.


"""
__author__ = """Aric Hagberg (hagberg@lanl.gov)\nPieter Swart (swart@lanl.gov)\nDan Schult(dschult@colgate.edu)"""
#    Copyright (C) 2004-2008 by
#    Aric Hagberg <hagberg@lanl.gov>
#    Dan Schult <dschult@colgate.edu>
#    Pieter Swart <swart@lanl.gov>
#    Distributed under the terms of the GNU Lesser General Public License
#    http://www.gnu.org/copyleft/lesser.html
#
__docformat__ = "restructuredtext en"


from pylispng.exception import GraphError


def empty_graph(n=0,create_using=None):
    """Return the empty graph with n nodes and zero edges.

    Node labels are the integers 0 to n-1

    For example:
    >>> G = empty_graph(10)
    >>> G.number_of_nodes()
    10
    >>> G.number_of_edges()
    0

    The variable create_using should point to a "graph"-like object that
    will be cleaned (nodes and edges will be removed) and refitted as
    an empty "graph" with n nodes with integer labels. This capability
    is useful for specifying the class-nature of the resulting empty
    "graph" (i.e. Graph, DiGraph, MyWeirdGraphClass, etc.).

    The variable create_using has two main uses:
    Firstly, the variable create_using can be used to create an
    empty digraph, network,etc.  For example,

    >>> n=10
    >>> G = empty_graph(n,create_using=DiGraph())

    will create an empty digraph on n nodes.

    Secondly, one can pass an existing graph (digraph, pseudograph,
    etc.) via create_using. For example, if G is an existing graph
    (resp. digraph, pseudograph, etc.), then empty_graph(n,create_using=G)
    will empty G (i.e. delete all nodes and edges using G.clear() in
    base) and then add n nodes and zero edges, and return the modified
    graph (resp. digraph, pseudograph, etc.).

    See also create_empty_copy(G).

    """
    if create_using is None:
        # default empty graph is a simple graph
        G = Graph()
    else:
        G = create_using
        G.clear()

    G.add_nodes_from(xrange(n))
    G.name="empty_graph(%d)"%n
    return G


def complete_graph(n,create_using=None):
    """ Return the Complete graph K_n with n nodes.

    Node labels are the integers 0 to n-1.

    """
    G = empty_graph(n,create_using)
    G.name="complete_graph(%d)"%n
    for u in xrange(n):
        for v in xrange(u+1,n):
            G.add_edge(u,v)
    return G


def path_graph(n,create_using=None):
    """Return the Path graph P_n of n nodes linearly connected
    by n-1 edges.

    Node labels are the integers 0 to n - 1.
    If create_using is a DiGraph then the edges are directed in
    increasing order.

    """
    G = empty_graph(n,create_using)
    G.name="path_graph(%d)"%n
    G.add_edges_from([(v,v+1) for v in xrange(n-1)])
    return G


class Graph(object):
    """
    An undirected graph class without multiple (parallel) edges.

    Nodes can be arbitrary (hashable) objects.

    Arbitrary data/labels/objects can be associated with edges.
    The default data is 1.  See add_edge and add_edges_from methods
    for details.  Many NetworkX routines developed for weighted graphs
    assume this data is a number.  Feel free to put other objects on
    the edges, but be aware that these weighted graph algorithms may
    give unpredictable results if the graph isn't a weighted graph.

    Self loops are allowed.

    Examples
    ========

    Create an empty graph structure (a "null graph") with no nodes and
    no edges.

    >>> from pylispng.graph import Graph
    >>> G = Graph()

    G can be grown in several ways.
    By adding one node at a time:

    >>> G.add_node(1)

    by adding a list of nodes:

    >>> G.add_nodes_from([2,3])

    by using an iterator:

    >>> G.add_nodes_from(xrange(100,110))

    or by adding any container of nodes (a list, dict, set
    or even a file or the nodes from another graph).

    >>> H = path_graph(10)
    >>> G.add_nodes_from(H)

    Any hashable object (except None) can represent a node,
    e.g. a customized node object, or even another Graph.

    >>> G.add_node(H)

    G can also be grown by adding one edge at a time:

    >>> G.add_edge(1, 2)

    by adding a list of edges:

    >>> G.add_edges_from([(1,2),(1,3)])

    or by adding any collection of edges:

    >>> G.add_edges_from(H.edges())

    Nodes will be added as needed when you add edges and there are
    no complaints when adding existing nodes or edges.

    The default edge data is the number 1.
    To add edge information with an edge, use a 3-tuple (u,v,d).

    >>> G.add_edges_from([(1,2,'blue'),(2,3,3.1)])
    >>> G.add_edges_from( [(3,4),(4,5)], data='red')
    >>> G.add_edge( 1, 2, 4.7 )

    """
    multigraph = False
    directed = False
    def __init__(self, name='', weighted=True):
        """Initialize an empty graph.

        Examples
        --------
        >>> G = Graph()
        >>> G = Graph(name='my graph')
        >>> G = Graph(weighted=False)  # don't assume edge data are weights
        """
        self.adj = {}  # empty adjacency dict
        self.weighted = weighted
        self.name = name

    def __str__(self):
        """Return the name."""
        return self.name

    def __iter__(self):
        """Iterate over the nodes. Use "for n in G".

        Examples
        --------
        >>> G = path_graph(4)
        >>> print [n for n in G]
        [0, 1, 2, 3]

        """
        return self.adj.iterkeys()

    def __contains__(self,n):
        """Return True if n is a node, False otherwise. Use "n in G".

        Examples
        --------
        >>> G = path_graph(4)
        >>> print 1 in G
        True
        """
        try:
            return n in self.adj
        except TypeError:
            return False

    def __len__(self):
        """Return the number of nodes. Use "len(G)".

        Examples
        --------
        >>> G = path_graph(4)
        >>> print len(G)
        4

        """
        return len(self.adj)

    def __getitem__(self, n):
        """Return the neighbors of node n.  Use "G[n]".

        Examples
        --------
        >>> G = path_graph(4)
        >>> print G[0]
        {1: 1}

        Notes
        -----
        G[n] is similar to G.neighbors(n) but the internal data dictionary
        is returned instead of a list.

        G[u][v] returns the edge data for edge (u,v).

        >>> G = path_graph(4)
        >>> print G[0][1]
        1

        Assigning G[u][v] may corrupt the graph data structure.
        """
        return self.adj[n]


    def add_node(self, n):
        """Add a single node n.

        Parameters
        ----------
        n : node
            A node n can be any hashable Python object except None.

            A hashable object is one that can be used as a key in a Python
            dictionary. This includes strings, numbers, tuples of strings
            and numbers, etc.

        Examples
        --------
        >>> G = Graph()
        >>> G.add_node(1)
        >>> G.add_node('Hello')
        >>> K3=complete_graph(3)
        >>> G.add_node(K3)
        >>> G.number_of_nodes()
        3

        Notes
        -----
        On many platforms hashable items also include mutables such as
        Graphs, though one should be careful that the hash doesn't
        change on mutables.

        """
        if n not in self.adj:
            self.adj[n] = {}


    def add_nodes_from(self, nbunch):
        """Add nodes from nbunch.

        Parameters
        ----------
        nbunch : list, iterable
            A container of nodes that will be iterated through once
            (thus it should be an iterator or be iterable).  Each
            element of the container should be a valid node type: any
            hashable type except None.

        Examples
        --------
        >>> G = Graph()
        >>> G.add_nodes_from('Hello')
        >>> K3=complete_graph(3)
        >>> G.add_nodes_from(K3)
        >>> sorted(G.nodes())
        [0, 1, 2, 'H', 'e', 'l', 'o']

        """
        for n in nbunch:
            if n not in self.adj:
                self.adj[n]={}


    def remove_node(self,n):
        """Remove node n.

        Removes the node n and adjacent edges in the graph.
        Attempting to remove a non-existent node will raise an exception.

        Examples
        --------
        >>> G = complete_graph(3)  # complete graph on 3 nodes, K3
        >>> G.edges()
        [(0, 1), (0, 2), (1, 2)]
        >>> G.remove_node(1)
        >>> G.edges()
        [(0, 2)]

        """
        adj = self.adj
        try:
            nbrs = adj[n].keys() # keys handles self-loops (allow mutation later)
        except KeyError: # GraphError if n not in self
            raise GraphError("node %s not in graph"%(n,))
        for u in nbrs:
            del adj[u][n]   # remove all edges n-u in graph
        del adj[n]          # now remove node


    delete_node = remove_node

    def remove_nodes_from(self, nbunch):
        """Remove nodes specified in nbunch.

        Parameters
        ----------
        nbunch : list, iterable
            A container of nodes that will be iterated through once
            (thus it should be an iterator or be iterable).  Each
            element of the container should be a valid node type: any
            hashable type except None.

        Examples
        --------
        >>> G = complete_graph(3)  # complete graph on 3 nodes, K3
        >>> e=G.nodes()
        >>> e
        [0, 1, 2]
        >>> G.remove_nodes_from(e)
        >>> G.nodes()
        []

        """
        adj = self.adj
        for n in nbunch:
            try:
                for u in adj[n].keys():   # keys() handles self-loops
                    del adj[u][n]         #(allows mutation of dict in loop)
                del adj[n]
            except KeyError:
                pass

    delete_nodes_from = remove_nodes_from

    def nodes_iter(self):
        """Return an iterator for the nodes.

        Examples
        --------
        >>> G = path_graph(3)
        >>> for n in G.nodes_iter():
        ...     print n,
        0 1 2

        You can also say

        >>> G = path_graph(3)
        >>> for n in G:
        ...     print n,
        0 1 2

        Notes
        -----
        It is simpler and equivalent to use the expression "for n in G"

        >>> G = path_graph(3)
        >>> for n in G:
        ...     print n,
        0 1 2

        """
        return self.adj.iterkeys()

    def nodes(self):
        """Return a list of the nodes.

        Examples
        --------
        >>> G = path_graph(3)
        >>> print G.nodes()
        [0, 1, 2]

        """
        return self.adj.keys()

    def number_of_nodes(self):
        """Return the number of nodes.

        Examples
        --------
        >>> G = path_graph(4)
        >>> print len(G)
        4

        Notes
        -----
        This is the same as

        >>> len(G)
        4

        and

        >>> G.order()
        4

        """
        return len(self.adj)

    def order(self):
        """Return the number of nodes.

        See Also
        --------
        Graph.order(), Graph.__len__()


        """
        return len(self.adj)

    def has_node(self,n):
        """Return True if graph has node n.

        Examples
        --------
        >>> G = path_graph(4)
        >>> print G.has_node(0)
        True

        Notes
        -----

        It is more readable and simpler to use
        >>> 0 in G
        True

        """
        try:
            return n in self.adj
        except TypeError:
            return False

    def rename_node(self, old, new, attr='adj'):
        """
        >>> g = Graph()
        >>> g.add_nodes_from([0, 1, 2, 3, 4, 5, 6])
        >>> g.add_edges_from([
        ...   (0, 1), (0, 2), (0, 3), (0, 6),
        ...   (3, 4), (3, 5)])
        >>> g.nodes()
        [0, 1, 2, 3, 4, 5, 6]
        >>> g.edges()
        [(0, 1), (0, 2), (0, 3), (0, 6), (3, 4), (3, 5)]

        >>> g.rename_node(4, 37)
        >>> g.rename_node(3, 52)
        >>> g.rename_node(0, 10)

        # XXX in the following, why is the '52' showing up in the middle and
        # not at the end?
        >>> g.nodes()
        [1, 2, 5, 6, 10, 52, 37]

        >>> g.edges()
        [(1, 10), (2, 10), (5, 52), (6, 10), (10, 52), (52, 37)]
        """
        mutateAttr = getattr(self, attr)
        for nodeID, relation in mutateAttr.items():
            for relationNodeID, data in relation.items():
                if relationNodeID == old:
                    mutateAttr[nodeID][new] = mutateAttr[nodeID][old]
                    del mutateAttr[nodeID][old]
            if nodeID == old:
                mutateAttr[new] = mutateAttr[old]
                del mutateAttr[old]

    def add_edge(self, u, v, data=1):
        """Add an edge between u and v with optional data.

        The nodes u and v will be automatically added if they are
        not already in the graph.

        Parameters
        ----------
        u,v : nodes
            Nodes can be, for example, strings or numbers.
            Nodes must be hashable (and not None) Python objects.

        data : Python object
            Edge data (or labels or objects) can be entered via the
            optional argument data which defaults to 1.

            Some NetworkX algorithms are designed for weighted
            graphs for which the edge data must be a number. These
            may behave unpredictably for edge data that isn't a number.

        Examples
        --------
        The following all add the edge e=(1,2) to graph G.

        >>> G = Graph()
        >>> e=(1,2)
        >>> G.add_edge( 1, 2 )          # explicit two node form
        >>> G.add_edge( *e)             # single edge as tuple of two nodes
        >>> G.add_edges_from( [(1,2)] ) # add edges from iterable container

        Associate the data myedge to the edge (1,2).

        >>> myedge=1.3
        >>> G.add_edge(1, 2, myedge)

        Notes
        -----
        Adding an edge that already
        exists *overwrites* the edgedata.

        See Also
        --------
        Parallel edges are allowed in the MultiGraph class.
        """
        # add nodes
        if u not in self.adj:
            self.adj[u] = {}
        if v not in self.adj:
            self.adj[v] = {}
        # add the edge
        self.adj[u][v] = data
        self.adj[v][u] = data


    def add_edges_from(self, ebunch, data=1):
        """Add all the edges in ebunch.

        Parameters
        ----------
        ebunch : list or container of edges
            The container must be iterable or an iterator.  It is
            iterated over once. Adding the same edge twice has no
            effect and does not raise an exception.  The edges in
            ebunch must be 2-tuples (u,v) or 3-tuples (u,v,d).

         data : any Python object
             The default data for edges with no data given.
             If unspecified the integer 1 will be used.

        See Also
        --------
        add_edge

        Examples
        --------
        >>> G = Graph()
        >>> G.add_edges_from([(0,1),(1,2)]) # using a list of edge tuples
        >>> e=zip(range(0,3),range(1,4))
        >>> G.add_edges_from(e) # Add the path graph 0-1-2-3

        """
        for e in ebunch:
            ne=len(e)
            if ne==3:
                u,v,d = e
            elif ne==2:
                u,v = e
                d = data
            else:
                raise GraphError,"Edge tuple %s must be a 2-tuple or 3-tuple."%(e,)
            if u not in self.adj:
                self.adj[u] = {}
            if v not in self.adj:
                self.adj[v] = {}
            self.adj[u][v] = d
            self.adj[v][u] = d


    def remove_edge(self, u, v):
        """Remove the edge between (u,v).

        Parameters
        ----------
        u,v: nodes


        See Also
        --------
        remove_edges_from : remove a collection of edges

        Examples
        --------
        >>> G = path_graph(4)
        >>> G.remove_edge(0,1)
        >>> e=(1,2)
        >>> G.remove_edge(*e) # unpacks e from an edge tuple
        >>> e=(2,3,'data')
        >>> G.remove_edge(*e[:2]) # edge tuple with data

        """
        try:
            del self.adj[u][v]
            if u != v:  # self loop needs only one entry removed
                del self.adj[v][u]
        except KeyError:
            raise GraphError("edge %s-%s not in graph"%(u,v))


    delete_edge=remove_edge


    def remove_edges_from(self, ebunch):
        """Remove all edges specified in ebunch.

        Parameters
        ----------
        ebunch: list or container of edge tuples
            A container of edge 2-tuples (u,v) or edge 3-tuples(u,v,d)
            though d is ignored unless we are a multigraph.

        See Also
        --------
        remove_edge

        Examples
        --------
        >>> G = path_graph(4)
        >>> ebunch=[(1,2),(2,3)]
        >>> G.remove_edges_from(ebunch)

        Notes
        -----
        Will fail silently if the edge (u,v) is not in the graph.

        """
        for e in ebunch:
            u,v = e[:2]  # ignore edge data if present
            if u in self.adj and v in self.adj[u]:
                del self.adj[u][v]
                if u != v:  # self loop needs only one entry deleted
                    del self.adj[v][u]


    delete_edges_from = remove_edges_from

    def has_neighbor(self, u, v):
        """Return True if node u has neighbor v.

        This returns True if there exists any edge (u,v,data) for some data.

        See Also
        --------
        has_edge

        Examples
        --------
        >>> G = path_graph(4)
        >>> G.has_neighbor(0,1)
        True
        >>> G.has_edge(0,1) # same as has_neighbor
        True
        >>> 1 in G[0]  # this gives KeyError if u not in G
        True

        """
        try:
            return v in self.adj[u]
        except KeyError:
            return False

    def has_edge(self, u, v):
        """Return True if graph contains the edge (u,v), False otherwise.

        See Also
        --------
        Graph.has_neighbor()

        Examples
        --------
        Can be called either using two nodes u,v or edge tuple (u,v)

        >>> G = path_graph(4)
        >>> G.has_edge(0,1)  # called using two nodes
        True
        >>> e=(0,1)
        >>> G.has_edge(*e)  #  e is a 2-tuple (u,v)
        True
        >>> e=(0,1,'data')
        >>> G.has_edge(*e[:2])  # e is a 3-tuple (u,v,d)
        True

        The following syntax are all equivalent:

        >>> G.has_neighbor(0,1)
        True
        >>> G.has_edge(0,1)
        True
        >>> 1 in G[0]  # though this gives KeyError if 0 not in G
        True

        """
        try:
            return v in self.adj[u]
        except KeyError:
            return False


    def neighbors(self, n):
        """Return a list of the nodes connected to the node n.

        Examples
        --------
        >>> G = path_graph(4)
        >>> G.neighbors(0)
        [1]

        Notes
        -----
        It is sometimes more convenient (and faster) to access
        the adjacency dictionary as G[n]

        >>> G = Graph()
        >>> G.add_edge('a','b','data')
        >>> G['a']
        {'b': 'data'}
        """
        try:
            return self.adj[n].keys()
        except KeyError:
            raise GraphError("node %s not in graph"%(n,))

    def neighbors_iter(self, n):
        """Return an iterator over all neighbors of node n.

        Examples
        --------
        >>> G = path_graph(4)
        >>> print [n for n in G.neighbors(0)]
        [1]

        Notes
        -----
        It is faster to iterate over the using the idiom
        >>> print [n for n in G[0]]
        [1]

        """
        try:
            return self.adj[n].iterkeys()
        except KeyError:
            raise GraphError("node %s not in graph"%(n,))

    def edges(self, nbunch=None, data=False):
        """Return a list of edges.

        Parameters
        ----------
        nbunch : list, iterable
            A container of nodes that will be iterated through once
            (thus it should be an iterator or be iterable).  Each
            element of the container should be a valid node type: any
            hashable type except None.
            If nbunch is None, return all edges in the graph.
            Nodes in nbunch that are not in the graph will be (quietly) ignored.

        data : bool
            Return two tuples (u,v) (False) or three-tuples (u,v,data) (True)


        Returns
        --------
        Edges that are adjacent to any node in nbunch,
        or a list of all edges if nbunch is not specified.

        Examples
        --------
        >>> G = path_graph(4)
        >>> G.edges()
        [(0, 1), (1, 2), (2, 3)]
        >>> G.edges(data=True) # default edge data is 1
        [(0, 1, 1), (1, 2, 1), (2, 3, 1)]
        """
        return list(self.edges_iter(nbunch, data))

    def edges_iter(self, nbunch=None, data=False):
        """Return an iterator over the edges.

        Parameters
        ----------
        nbunch : list, iterable
            A container of nodes that will be iterated through once
            (thus it should be an iterator or be iterable).  Each
            element of the container should be a valid node type: any
            hashable type except None.
            If nbunch is None, return all edges in the graph.
            Nodes in nbunch that are not in the graph will be (quietly) ignored.

        data : bool
            Return two tuples (u,v) (False) or three-tuples (u,v,data) (True)

        Returns
        -------
        An iterator over edges that are adjacent to any node in nbunch,
        or over all edges if nbunch is not specified.

        Examples
        --------
        >>> G = path_graph(4)
        >>> G.edges()
        [(0, 1), (1, 2), (2, 3)]
        >>> G.edges(data=True) # default edge data is 1
        [(0, 1, 1), (1, 2, 1), (2, 3, 1)]
        """
        seen={}     # helper dict to keep track of multiply stored edges
        if nbunch is None:
            nodes_nbrs = self.adj.iteritems()
        else:
            nodes_nbrs=((n,self.adj[n]) for n in self.nbunch_iter(nbunch))
        if data:
            for n,nbrs in nodes_nbrs:
                for nbr,data in nbrs.iteritems():
                    if nbr not in seen:
                        yield (n,nbr,data)
                seen[n]=1
        else:
            for n,nbrs in nodes_nbrs:
                for nbr in nbrs:
                    if nbr not in seen:
                        yield (n,nbr)
                seen[n] = 1
        del seen


    def get_edge(self, u, v, default=None):
        """Return the data associated with the edge (u,v).

        Parameters
        ----------
        u,v : nodes
            If u or v are not nodes in graph an exception is raised.

        default:  any Python object
            Value to return if the edge (u,v) is not found.
            If not specified, raise an exception.

        Examples
        --------
        >>> G = path_graph(4) # path graph with edge data all set to 1
        >>> G.get_edge(0,1)
        1
        >>> e=(0,1)
        >>> G.get_edge(*e) # tuple form
        1

        Notes
        -----
        It is faster to use G[u][v].

        >>> G[0][1]
        1

        """
        try:
            return self.adj[u][v]
        except KeyError:
            if default is not None and u in self and v in self: return default
            raise GraphError("edge (%s,%s) not in graph"%(u,v))


    def adjacency_list(self):
        """Return an adjacency list as a Python list of lists

        The output adjacency list is in the order of G.nodes().
        For directed graphs, only outgoing adjacencies are included.

        Examples
        --------
        >>> G = path_graph(4)
        >>> G.adjacency_list() # in sorted node order 0,1,2,3
        [[1], [0, 2], [1, 3], [2]]

        """
        return map(list,self.adj.itervalues())

    def adjacency_iter(self):
        """Return an iterator of (node, adjacency dict) tuples for all
        nodes.

        This is the fastest way to look at every edge.
        For directed graphs, only outgoing adjacencies are included.

        Examples
        --------
        >>> G = path_graph(4)
        >>> [(n,nbrdict) for n,nbrdict in G.adjacency_iter()]
        [(0, {1: 1}), (1, {0: 1, 2: 1}), (2, {1: 1, 3: 1}), (3, {2: 1})]

        Notes
        -----
        The dictionary returned is part of the internal graph data structure;
        changing it could corrupt that structure.
        This is meant for fast inspection, not mutation.

        For MultiGraph/MultiDiGraph multigraphs, a list of edge
        data is the value in the dict.

        """
        return self.adj.iteritems()

    def degree(self, nbunch=[], with_labels=False, weighted=False):
        """Return the degree of a node or nodes.

        The node degree is the number of edges adjacent to that node.

        Parameters
        ----------
        nbunch : list, iterable
            A container of nodes that will be iterated through once
            (thus it should be an iterator or be iterable).  Each
            element of the container should be a valid node type: any
            hashable type except None.
            If nbunch is None, return all edges data in the graph.
            Nodes in nbunch that are not in the graph will be (quietly) ignored.

        with_labels : False|True
            Return a list of degrees (False) or a dictionary of
            degrees keyed by node (True).

        weighted : False|True
            If the graph is weighted return the weighted degree
            (the sum of edge weights).

        Examples
        --------
        >>> G = path_graph(4)
        >>> G.degree(0)
        [1]
        >>> G.degree([0,1])
        [1, 2]
        >>> G.degree([0,1],with_labels=True)
        {0: 1, 1: 2}

        """
        try:
            iter(nbunch)
        except TypeError:
            nbunch = [nbunch]
        if with_labels:           # return a dict
            return dict(self.degree_iter(nbunch, weighted=weighted))
        elif nbunch in self:      # return a single node
            return self.degree_iter(nbunch, weighted=weighted).next()[1]
        else:                     # return a list
            return [d for (n,d) in self.degree_iter(nbunch, weighted=weighted)]

    def degree_iter(self, nbunch=[], weighted=False):
        """Return an iterator for (node, degree).

        The node degree is the number of edges adjacent to that node.

        Parameters
        ----------
        nbunch : list, iterable
            A container of nodes that will be iterated through once
            (thus it should be an iterator or be iterable).  Each
            element of the container should be a valid node type: any
            hashable type except None.
            If nbunch is None, return all edges data in the graph.
            Nodes in nbunch that are not in the graph will be (quietly) ignored.

        weighted : False|True
            If the graph is weighted return the weighted degree
            (the sum of edge weights).

        Examples
        --------
        >>> G = path_graph(4)
        >>> list(G.degree_iter(0)) # node 0 with degree 1
        [(0, 1)]
        >>> list(G.degree_iter([0,1]))
        [(0, 1), (1, 2)]
        """
        if nbunch == []:
            nodes_nbrs = self.adj.iteritems()
        else:
            nodes_nbrs=((n,self.adj[n]) for n in self.nbunch_iter(nbunch))

        if self.weighted and weighted:
        # edge weighted graph - degree is sum of nbr edge weights
            for n,nbrs in nodes_nbrs:
                yield (n,sum(nbrs.itervalues())+(n in nbrs and nbrs[n]))
        else:
            for n,nbrs in nodes_nbrs:
                yield (n,len(nbrs)+(n in nbrs)) # return tuple (n,degree)

    def clear(self):
        """Remove all nodes and edges.

        This also removes the name.

        Examples
        --------
        >>> G = path_graph(4)
        >>> G.clear()
        >>> G.nodes()
        []
        >>> G.edges()
        []

        """
        self.name = ''
        self.adj.clear()

    def copy(self):
        """Return a copy of the graph.

        Notes
        -----
        This makes a complete of the graph but does not make copies
        of any underlying node or edge data.  The node and edge data
        in the copy still point to the same objects as in the original.
        """
        H = self.__class__()
        H.name = self.name
        H.weighted = self.weighted
        H.adj = self.adj
        return H

    def subgraph(self, nbunch, copy=True):
        """Return the subgraph induced on nodes in nbunch.

        Parameters
        ----------
        nbunch : list, iterable
            A container of nodes that will be iterated through once
            (thus it should be an iterator or be iterable).  Each
            element of the container should be a valid node type: any
            hashable type except None.
            If nbunch is None, return all edges data in the graph.
            Nodes in nbunch that are not in the graph will be (quietly) ignored.

        copy : bool (default True)
            If True return a new graph holding the subgraph.
            Otherwise, the subgraph is created in the original
            graph by deleting nodes not in nbunch.
            Warning: this can destroy the graph.
        """
        bunch =set(self.nbunch_iter(nbunch))

        if not copy:
            # demolish all nodes (and attached edges) not in nbunch
            self.remove_nodes_from([n for n in self if n not in bunch])
            self.name = "Subgraph of (%s)"%(self.name)
            return self
        else:
            # create new graph and copy subgraph into it
            H = self.__class__()
            H.name = "Subgraph of (%s)"%(self.name)
            # add edges
            H_adj = H.adj # cache
            self_adj = self.adj # cache
            for n in bunch:
                H_adj[n] = dict( ((u,d) for u,d in self_adj[n].iteritems()
                                 if u in bunch) )
            return H

    def nodes_with_selfloops(self):
        """Return a list of nodes with self loops.

        Examples
        --------
        >>> G = Graph()
        >>> G.add_edge(1,1)
        >>> G.add_edge(1,2)
        >>> G.nodes_with_selfloops()
        [1]
        """
        return [ n for n,nbrs in self.adj.iteritems() if n in nbrs ]

    def selfloop_edges(self,data=False):
        """Return a list of selfloop edges

        Parameters
        -----------
        data : bool
            Return two tuples (u,v) (False) or three-tuples (u,v,data) (True)

        Examples
        --------
        >>> G = Graph()
        >>> G.add_edge(1,1)
        >>> G.add_edge(1,2)
        >>> G.selfloop_edges()
        [(1, 1)]
        >>> G.selfloop_edges(data=True)
        [(1, 1, 1)]

        """
        if data:
            return [ (n,n,nbrs[n])
                     for n,nbrs in self.adj.iteritems() if n in nbrs ]
        else:
            return [ (n,n)
                     for n,nbrs in self.adj.iteritems() if n in nbrs ]

    def number_of_selfloops(self):
        """Return the number of selfloop edges (edge from a node to
        itself).

        Examples
        --------
        >>> G = Graph()
        >>> G.add_edge(1,1)
        >>> G.add_edge(1,2)
        >>> G.number_of_selfloops()
        1
        """
        return len(self.nodes_with_selfloops())

    def size(self, weighted=False):
        """Return the number of edges.

        Parameters
        ----------
        weighted : bool, optional
           If True return the sum of the edge weights.

        Examples
        --------
        >>> G = path_graph(4)
        >>> G.size()
        3

        >>> G = Graph()
        >>> G.add_edge('a','b',2)
        >>> G.add_edge('b','c',4)
        >>> G.size()
        2
        >>> G.size(weighted=True)
        6

        See Also
        --------
        Graph.number_of_edges()
        """
        return sum(self.degree(weighted=weighted))/2

    def number_of_edges(self, u=None, v=None):
        """Return the number of edges between two nodes.

        Parameters
        ----------
        u,v : nodes
            If u and v are specified, return the number of edges between
            u and v. Otherwise return the total number of all edges.

        Examples
        --------
        >>> G = path_graph(4)
        >>> G.number_of_edges()
        3
        >>> G.number_of_edges(0,1)
        1
        >>> e=(0,1)
        >>> G.number_of_edges(*e)
        1
        """
        if u is None: return self.size()
        if v in self.adj[u]:
            return 1
        else:
            return 0

    def is_directed(self):
        return self.directed

    def add_star(self, nlist, data=None):
        """Add a star.

        The first node in nlist is the middle of the star.  It is connected
        to all other nodes in nlist.

        Parameters
        ----------
        nlist : list
            A list of nodes.

        data : list or iterable, optional
            Data to add to the edges in the path.
            The length should be one less than len(nlist).

        Examples
        --------
        >>> G = Graph()
        >>> G.add_star([0,1,2,3])
        >>> G.add_star([10,11,12],data=['a','b'])
        """
        if data is None:
            v=nlist[0]
            edges=((v,n) for n in nlist[1:])
        else:
            # if len(data)<len(nlist) only len(data) items will be added
            v=nlist[0]
            edges=((v,n,d) for n,d in zip(nlist[1:],data))
        self.add_edges_from(edges)

    def add_path(self, nlist, data=None):
        """Add a path.

        Parameters
        ----------
        nlist : list
            A list of nodes.  A path will be constructed from
            the nodes (in order) and added to the graph.

        data : list or iterable, optional
            Data to add to the edges in the path.
            The length should be one less than len(nlist).

        Examples
        --------
        >>> G = Graph()
        >>> G.add_path([0,1,2,3])
        >>> G.add_path([10,11,12],data=['a','b'])

        """
        if data is None:
            edges=zip(nlist[:-1],nlist[1:])
        else:
            # if len(data)<len(nlist) only len(data) items will be added
            edges=zip(nlist[:-1],nlist[1:],data)
        self.add_edges_from(edges)

    def add_cycle(self, nlist, data=None):
        """Add a cycle.

        Parameters
        ----------
        nlist : list
            A list of nodes.  A cycle will be constructed from
            the nodes (in order) and added to the graph.

        data : list or iterable, optional
            Data to add to the edges in the path.
            The length should be the same as nlist.

        Examples
        --------
        >>> G = Graph()
        >>> G.add_cycle([0,1,2,3])
        >>> G.add_cycle([10,11,12],data=['a','b','c'])
        """
        if data is None:
            edges=zip(nlist,nlist[1:]+[nlist[0]])
        else:
            # if len(data)<len(nlist) only len(data) items will be added
            edges=zip(nlist,nlist[1:]+[nlist[0]],data)
        self.add_edges_from(edges)

    def nbunch_iter(self, nbunch=None):
        """Return an iterator of nodes contained
        in nbunch that are also in the graph.

        Parameters
        ----------
        nbunch : list, iterable
            A container of nodes that will be iterated through once
            (thus it should be an iterator or be iterable).  Each
            element of the container should be a valid node type: any
            hashable type except None.
            If nbunch is None, return all edges data in the graph.
            Nodes in nbunch that are not in the graph will be (quietly) ignored.

        Notes
        -----
        When nbunch is an iterator, the returned iterator yields values
        directly from nbunch, becoming exhausted when nbunch is exhausted.

        To test whether nbunch is a single node, one can use
        "if nbunch in self:", even after processing with this routine.

        If nbunch is not a node or a (possibly empty) sequence/iterator
        or None, a GraphError is raised.  Also, if any values returned
        by an iterator nbunch is not hashable, a GraphError is raised.
        """
        # include all nodes via iterator
        if nbunch == None:
            bunch=self.adj.iterkeys()
        # if nbunch is a single node
        elif not hasattr(nbunch, '__iter__') and nbunch in self:
            bunch=[nbunch].__iter__()
        # if nbunch is a sequence of nodes
        else:
            def bunch_iter(nlist, adj):
                if not hasattr(nlist, '__iter__'):
                    nlist = [nlist]
                try:
                    for n in nlist:
                        if n in adj:
                            yield n
                except TypeError,e:
                    # capture error for non-sequence/iterator nbunch.
                    if 'iterable' in e.message:
                        raise GraphError(
                            "nbunch is not a node or a sequence of nodes.")
                    # capture error for unhashable node.
                    elif 'hashable' in e.message:
                        raise GraphError(
                            "Node %s in the sequence nbunch is not a valid node."%n)
                    else:
                        raise
            bunch=bunch_iter(nbunch,self.adj)
        return bunch


class DiGraph(Graph):
    """ A directed graph that allows self-loops, but not multiple
    (parallel) edges.

    Edge and node data is the same as for Graph.
    Subclass of Graph.

    An empty digraph is created with

    >>> G = DiGraph()

    """
    multigraph=False
    directed=True
    def __init__(self, name='', weighted=True):
        """Initialize an empty directed graph.

        Examples
        --------
        >>> G = DiGraph()
        >>> G = DiGraph(name='my graph')

        # don't assume edge data are weights
        >>> G = DiGraph(weighted=False)
        """
        # We store two adjacency lists:
        # the  predecessors of node n are stored in the dict self.pred
        # the successors of node n are stored in the dict self.succ=self.adj
        self.adj = {}  # empty adjacency dictionary
        self.pred = {}  # predecessor
        self.succ = self.adj  # successor
        self.weighted = weighted
        self.name=name


    def add_node(self, n):
        if n not in self.succ:
            self.succ[n] = {}
            self.pred[n] = {}

    add_node.__doc__ = Graph.add_node.__doc__.replace('Graph','DiGraph')

    def add_nodes_from(self, nbunch):
        for n in nbunch:
            if n not in self.succ:
                self.add_node(n)

    add_nodes_from.__doc__ = \
        Graph.add_nodes_from.__doc__.replace('Graph','DiGraph')


    def remove_node(self, n):
        try:
            nbrs=self.succ[n]
        except KeyError: # GraphError if n not in self
            raise GraphError, "node %s not in digraph"%(n,)
        for u in nbrs:
            del self.pred[u][n] # remove all edges n-u in digraph
        del self.succ[n]          # remove node from succ
        for u in self.pred[n]:
            del self.succ[u][n] # remove all edges n-u in digraph
        del self.pred[n]          # remove node from pred

    remove_node.__doc__ = Graph.remove_node.__doc__.replace('Graph','DiGraph')
    delete_node = remove_node

    def remove_nodes_from(self, nbunch):
        for n in nbunch:
            try:
                succs=self.succ[n]
                for u in succs:
                    del self.pred[u][n] # remove all edges n-u in digraph
                del self.succ[n]          # now remove node
                for u in self.pred[n]:
                    del self.succ[u][n] # remove all edges n-u in digraph
                del self.pred[n]          # now remove node
            except KeyError:
                pass # silent failure on remove

    remove_nodes_from.__doc__ = \
        Graph.remove_nodes_from.__doc__.replace('Graph','DiGraph')
    delete_nodes_from = remove_nodes_from

    def rename_node(self, old, new):
        """
        >>> d = DiGraph()
        >>> d.add_nodes_from([0, 1, 2, 3, 4, 5, 6, 7])
        >>> d.add_edges_from([
        ...   (0, 1), (0, 2), (0, 3), (0, 6),
        ...   (3, 4), (3, 5),
        ...   (7, 0)])
        >>> d.nodes()
        [0, 1, 2, 3, 4, 5, 6, 7]
        >>> d.edges()
        [(0, 1), (0, 2), (0, 3), (0, 6), (3, 4), (3, 5), (7, 0)]

        >>> d.rename_node(4, 37)
        >>> d.rename_node(3, 52)
        >>> d.rename_node(0, 10)

        >>> d.nodes()
        [1, 2, 5, 6, 7, 10, 52, 37]
        >>> d.edges()
        [(7, 10), (10, 1), (10, 2), (10, 52), (10, 6), (52, 5), (52, 37)]
        """
        super(DiGraph, self).rename_node(old, new)
        super(DiGraph, self).rename_node(old, new, attr="pred")

    def add_edge(self, u, v, data=1):
        # add nodes
        if u not in self.succ:
            self.add_node(u)
        if v not in self.succ:
            self.add_node(v)
        # add edge
        self.succ[u][v]=data
        self.pred[v][u]=data

    def add_edges_from(self, ebunch, data=1):
        for e in ebunch:
            ne = len(e)
            if ne==3:
                u,v,d = e
            elif ne==2:
                u,v = e
                d = data
            else:
                raise GraphError,"Edge tuple %s must be a 2-tuple or 3-tuple."%(e,)
            if u not in self.succ:
                self.succ[u] = {}
                self.pred[u] = {}
            if v not in self.succ:
                self.succ[v] = {}
                self.pred[v] = {}
            self.succ[u][v] = d
            self.pred[v][u] = d

    add_edges_from.__doc__ = \
        Graph.add_edges_from.__doc__.replace('Graph','DiGraph')

    def remove_edge(self, u, v):
        try:
            del self.succ[u][v]
            del self.pred[v][u]
        except KeyError:
            raise GraphError("edge %s-%s not in graph"%(u,v))

    remove_edge.__doc__ = Graph.remove_edge.__doc__.replace('Graph','DiGraph')
    delete_edge = remove_edge


    def remove_edges_from(self, ebunch):
        for e in ebunch:
            (u,v)=e[:2]  # ignore edge data
            if u in self.succ and v in self.succ[u]:
                del self.succ[u][v]
                del self.pred[v][u]

    remove_edges_from.__doc__ = \
        Graph.remove_edges_from.__doc__.replace('Graph','DiGraph')
    delete_edges_from = remove_edges_from


    def has_successor(self, u, v):
        """Return True if node u has successor v.

        This is true if graph has the edge u->v.
        """
        return (u in self.succ and v in self.succ[u])

    def has_predecessor(self, u, v):
        """Return True if node u has predecessor v.

        This is true if graph has the edge u<-v.
        """
        return (u in self.pred and v in self.pred[u])

    def successors_iter(self,n):
        """Return an iterator over successor nodes of n."""
        try:
            return self.succ[n].iterkeys()
        except KeyError:
            raise GraphError, "Node %s not in digraph"%(n,)

    def predecessors_iter(self,n):
        """Return an iterator over predecessor nodes of n."""
        try:
            return self.pred[n].iterkeys()
        except KeyError:
            raise GraphError, "node %s not in digraph"%(n,)

    def successors(self, n):
        """Return a list of successor nodes of n."""
        return list(self.successors_iter(n))

    def predecessors(self, n):
        """Return a list of predecessor nodes of n."""
        return list(self.predecessors_iter(n))


    # digraph definitions
    neighbors = successors
    neighbors.__doc__ = Graph.neighbors.__doc__
    neighbors_iter = successors_iter
    neighbors_iter.__doc__ = Graph.neighbors_iter.__doc__

    def edges_iter(self, nbunch=None, data=False):
        if nbunch is None:
            nodes_nbrs=self.adj.iteritems()
        else:
            nodes_nbrs=((n,self.adj[n]) for n in self.nbunch_iter(nbunch))
        if data:
            for n,nbrs in nodes_nbrs:
                for nbr,data in nbrs.iteritems():
                    yield (n,nbr,data)
        else:
            for n,nbrs in nodes_nbrs:
                for nbr in nbrs:
                    yield (n,nbr)

    edges_iter.__doc__ = Graph.edges_iter.__doc__


    def degree_iter(self, nbunch=[], weighted=False):
        if nbunch == []:
            nodes_nbrs=self.adj.iteritems()
        else:
            nodes_nbrs=((n,self.adj[n]) for n in self.nbunch_iter(nbunch))

        if self.weighted and weighted:
        # edge weighted graph - degree is sum of edge weights
            for n,nbrs in nodes_nbrs:
                yield (n,sum(nbrs.itervalues())+sum(self.pred[n].itervalues()))
        else:
            for n,nbrs in nodes_nbrs:
                yield (n,len(nbrs)+len(self.pred[n]))

    degree_iter.__doc__ = Graph.degree_iter.__doc__

    def in_degree_iter(self, nbunch=None, weighted=False):
        """Return an iterator for (node, in degree).

        The node in degree is the number of edges pointing in to that node.

        Parameters
        ----------
        nbunch : list or container of nodes
            Any single node or any sequence/iterator of nodes.
            If nbunch is None, return all edges data in the graph.
            Nodes in nbunch that are not in the graph will be (quietly) ignored.

        weighted : False|True
            If the graph is weighted return the weighted degree
            (the sum of edge weights).

        Examples
        ---------
        >>> G = DiGraph()
        >>> G.add_nodes_from([0,1,2])
        >>> G.add_edge(0,1)
        >>> G.add_edge(2,1)
        >>> G.add_edge(0,2)
        >>> G.add_edge(3,2)
        >>> G.in_degree_iter(0).next()
        (0, 0)
        >>> G.in_degree_iter(1).next()
        (1, 2)
        >>> G.in_degree_iter(2).next()
        (2, 2)
        >>> G.in_degree_iter(3).next()
        (3, 0)
        >>> list(G.in_degree_iter([0,1,2,3]))
        [(0, 0), (1, 2), (2, 2), (3, 0)]
        """
        if nbunch is None:
            nodes_nbrs=self.pred.iteritems()
        else:
            nodes_nbrs=((n,self.pred[n]) for n in self.nbunch_iter(nbunch))

        if self.weighted and weighted:
        # edge weighted graph - degree is sum of edge weights
            for n,nbrs in nodes_nbrs:
                yield (n,sum(nbrs.itervalues()))
        else:
            for n,nbrs in nodes_nbrs:
                yield (n,len(nbrs))


    def out_degree_iter(self, nbunch=None, weighted=False):
        """Return an iterator for (node, out degree).

        The node degree is the number of edges pointing out of that node.

        Parameters
        ----------
        nbunch : list or container of nodes
            Any single node or any sequence/iterator of nodes.
            If nbunch is None, return all edges data in the graph.
            Nodes in nbunch that are not in the graph will be (quietly) ignored.

        weighted : False|True
            If the graph is weighted return the weighted degree
            (the sum of edge weights).

        Examples
        ---------
        >>> G = DiGraph()
        >>> G.add_nodes_from([0,1,2])
        >>> G.add_edge(0,1)
        >>> G.add_edge(2,1)
        >>> G.add_edge(0,2)
        >>> G.add_edge(3,2)
        >>> G.out_degree_iter(0).next()
        (0, 2)
        >>> G.out_degree_iter(1).next()
        (1, 0)
        >>> G.out_degree_iter(2).next()
        (2, 1)
        >>> G.out_degree_iter(3).next()
        (3, 1)
        >>> list(G.out_degree_iter([0,1,2,3]))
        [(0, 2), (1, 0), (2, 1), (3, 1)]
        """
        if nbunch is None:
            nodes_nbrs=self.succ.iteritems()
        else:
            nodes_nbrs=((n,self.succ[n]) for n in self.nbunch_iter(nbunch))

        if self.weighted and weighted:
        # edge weighted graph - degree is sum of edge weights
            for n,nbrs in nodes_nbrs:
                yield (n,sum(nbrs.itervalues()))
        else:
            for n,nbrs in nodes_nbrs:
                yield (n,len(nbrs))


    def in_degree(self, nbunch=None, with_labels=False, weighted=False):
        """Return the in-degree of a node or nodes.

        The node in-degree is the number of edges pointing in to that node.

        Parameters
        ----------
        nbunch : list or container of nodes
            Any single node or any sequence/iterator of nodes.
            If nbunch is None, return all edges data in the graph.
            Nodes in nbunch that are not in the graph will be (quietly) ignored.

        with_labels : False|True
            Return a list of degrees (False) or a dictionary of
            degrees keyed by node (True).

        weighted : False|True
            If the graph is weighted return the weighted degree
            (the sum of edge weights).

        Examples
        ---------
        >>> G = DiGraph()
        >>> G.add_nodes_from([0,1,2])
        >>> G.add_edge(0,1)
        >>> G.add_edge(2,1)
        >>> G.add_edge(0,2)
        >>> G.add_edge(3,2)
        >>> G.in_degree(0)
        0
        >>> G.in_degree(1)
        2
        >>> G.in_degree(2)
        2
        >>> G.in_degree(3)
        0
        >>> G.in_degree([0,1,2,3])
        [0, 2, 2, 0]
        >>> G.in_degree([0,1,2,3],with_labels=True)
        {0: 0, 1: 2, 2: 2, 3: 0}

        """
        if with_labels:           # return a dict
            return dict(self.in_degree_iter(nbunch,weighted=weighted))
        elif nbunch in self:      # return a single node
            return self.in_degree_iter(nbunch,weighted=weighted).next()[1]
        else:                     # return a list
            return [d for (n,d) in self.in_degree_iter(nbunch,weighted=weighted)]


    def out_degree(self, nbunch=None, with_labels=False, weighted=False):
        """Return the in-degree of a node or nodes.

        The node in-degree is the number of edges pointing in to that node.

        Parameters
        ----------
        nbunch : list or container of nodes
            Any single node or any sequence/iterator of nodes.
            If nbunch is None, return all edges data in the graph.
            Nodes in nbunch that are not in the graph will be (quietly) ignored.

        with_labels : False|True
            Return a list of degrees (False) or a dictionary of
            degrees keyed by node (True).

        weighted : False|True
            If the graph is weighted return the weighted degree
            (the sum of edge weights).

        Examples
        ---------
        >>> G = DiGraph()
        >>> G.add_nodes_from([0,1,2])
        >>> G.add_edge(0,1)
        >>> G.add_edge(2,1)
        >>> G.add_edge(0,2)
        >>> G.add_edge(3,2)
        >>> G.out_degree(0)
        2
        >>> G.out_degree(1)
        0
        >>> G.out_degree(2)
        1
        >>> G.out_degree(3)
        1
        >>> G.out_degree([0,1,2,3])
        [2, 0, 1, 1]
        >>> G.out_degree([0,1,2,3],with_labels=True)
        {0: 2, 1: 0, 2: 1, 3: 1}

        """
        if with_labels:           # return a dict
            return dict(self.out_degree_iter(nbunch,weighted=weighted))
        elif nbunch in self:      # return a single node
            return self.out_degree_iter(nbunch,weighted=weighted).next()[1]
        else:                     # return a list
            return [d for (n,d) in self.out_degree_iter(nbunch,weighted=weighted)]

    def copy(self):
        H = super(DiGraph, self).copy()
        H.pred = self.pred
        H.succ = self.succ
        return H

    def clear(self):
        self.name=''
        self.succ.clear()
        self.pred.clear()

    def subgraph(self, nbunch, copy=True, reindex=False):
        """
        >>> d = DiGraph()
        >>> d.add_nodes_from([0, 1, 2, 3, 4, 5, 6])
        >>> d.add_edges_from([
        ...   (0, 1), (0, 2), (0, 3), (0, 6),
        ...   (3, 4), (3, 5)])
        >>> d.nodes()
        [0, 1, 2, 3, 4, 5, 6]
        >>> d.edges()
        [(0, 1), (0, 2), (0, 3), (0, 6), (3, 4), (3, 5)]

        >>> d2 = d.subgraph(d.nodes())
        >>> d2.nodes()
        [0, 1, 2, 3, 4, 5, 6]
        >>> d2.edges()
        [(0, 1), (0, 2), (0, 3), (0, 6), (3, 4), (3, 5)]

        >>> d3 = d.subgraph([3, 4, 5])
        >>> d3.nodes()
        [3, 4, 5]
        >>> d3.edges()
        [(3, 4), (3, 5)]

        >>> d3 = d.subgraph([3, 4, 5], reindex=True)
        >>> d3.nodes()
        [0, 1, 2]
        >>> d3.edges()
        [(0, 1), (0, 2)]
        """
        bunch = set(self.nbunch_iter(nbunch))
        if not copy:
            # demolish all nodes (and attached edges) not in nbunch
            self.remove_nodes_from([n for n in self if n not in bunch])
            self.name = "Subgraph of (%s)" % (self.name)
            return self
        else:
            if reindex:
                H = self.copy()
                # demolish all nodes (and attached edges) not in nbunch
                H.remove_nodes_from([n for n in H if n not in bunch])
                H.name = "Subgraph of (%s)" % (H.name)
                for new, old in enumerate(H.nodes()):
                    H.rename_node(old, new)
            else:
                # create new graph and copy subgraph into it
                H = self.__class__()
                H.name = "Subgraph of (%s)" % (self.name)
                # add edges
                H_succ = H.succ       # store in local variables
                H_pred = H.pred
                self_succ = self.succ
                self_pred = self.pred
                for n in bunch:
                    H_succ[n] = dict([(u,d) for u,d in self_succ[n].iteritems()
                                    if u in bunch])
                    H_pred[n] = dict([(u,d) for u,d in self_pred[n].iteritems()
                                    if u in bunch])
            return H

    def to_undirected(self):
        """Return an undirected representation of the digraph.

        A new graph is returned with the same name and nodes and
        with edge (u,v,data) if either (u,v,data) or (v,u,data)
        is in the digraph.  If both edges exist in digraph and
        their edge data is different, only one edge is created
        with an arbitrary choice of which edge data to use.
        You must check and correct for this manually if desired.

        """
        H = Graph()
        H.name=self.name
        H.add_nodes_from(self)
        H.add_edges_from([(v,u,d) for (u,v,d) in self.edges_iter(data=True)])
        return H

    def reverse(self, copy=True):
        """Return the reverse of the graph

        The reverse is a graph with the same nodes and edges
        but with the directions of the edges reversed.
        """
        H = self.__class__(name="Reverse of (%s)"%self.name)
        H.add_nodes_from(self)
        H.add_edges_from([(v,u,d) for (u,v,d) in self.edges_iter(data=True)])
        return H


class Vertex(object):
    """
    An object useful for maintaining identifiable objects in a graph.

    >>> Vertex(0, 1)
    <Vertex 0 (1)>
    >>> Vertex(1, 7.5)
    <Vertex 1 (7.5)>
    >>> Vertex(2, list)
    <Vertex 2 (<type 'list'>)>

    >>> class Add(object):
    ...   def __repr__(self):
    ...     return '+'

    >>> class Subtract(object):
    ...   def __repr__(self):
    ...     return '-'

    >>> Vertex(3, Add())
    <Vertex 3 (+)>
    >>> Vertex(4, Subtract())
    <Vertex 4 (-)>
    """
    def __init__(self, id, obj):
        self.id = id
        self.object = obj

    def __repr__(self):
        return "<%s %s (%s)>" % (
            self.__class__.__name__, self.id, (self.object))


class Tree(DiGraph):
    """
    This is not a tree in the rigorous, graph-theoretical sense. It's merely
    used as one.

    >>> g = Tree()
    >>> g.add_node(Vertex(0, '+'))
    >>> g.add_node(Vertex(1, 3))
    >>> g.add_edge(0, 1)
    >>> g.add_node(Vertex(2, 6))
    >>> g.add_edge(0, 2)
    >>> len(g)
    3

    >>> g[0]
    <Vertex 0 (+)>
    >>> g[1]
    <Vertex 1 (3)>
    >>> g[2]
    <Vertex 2 (6)>

    Now let's add an expression further down the tree and then look at the ins
    and outs:

    >>> g.add_node(Vertex(3, '*'))
    >>> g.add_edge(0, 3)
    >>> g.add_node(Vertex(4, 3))
    >>> g.add_edge(3, 4)
    >>> g.add_node(Vertex(5, 7))
    >>> g.add_edge(3, 5)
    >>> g.add_node(Vertex(6, 13))
    >>> g.add_edge(0, 6)

    >>> g.nodes()
    [0, 1, 2, 3, 4, 5, 6]
    >>> g.edges()
    [(0, 1), (0, 2), (0, 3), (0, 6), (3, 4), (3, 5)]
    >>> g.successors(3)
    [4, 5]
    >>> g.successors(0)
    [1, 2, 3, 6]
    """
    def __init__(self, *args, **kwds):
        super(Tree, self).__init__(*args, **kwds)
        self.vertexLookup = {}
        self.branchNodes = []
        self.leafNodes = []

    def __getitem__(self, key):
        """
        >>> g = Tree()
        >>> g.add_nodes_from([0, 1, 2, 3, 4, 5, 6])
        >>> g[0]
        <Vertex 0 (0)>
        >>> g[6]
        <Vertex 6 (6)>
        """
        return self.vertexLookup.get(key)

    def add_node(self, obj):
        """
        This method expects an object to be passed to it. This object and an
        automatically-assigned id are stored in a Vertex instance. If the
        passed object is already a vertex, it is simply stored as-is.

        >>> g = Tree()
        >>> g.add_node(1)
        >>> g[1]
        <Vertex 1 (1)>

        >>> g.add_node(Vertex(7, "+"))
        >>> g[7]
        <Vertex 7 (+)>
        """
        if not isinstance(obj, Vertex):
            obj = Vertex(obj, obj)
        if obj.id in self.vertexLookup.keys():
            msg = "An item with that index is already in the graph"
            raise GraphError, msg
        super(Tree, self).add_node(obj.id)
        self.vertexLookup.update({obj.id: obj})

    def add_nodes_from(self, nodes):
        for node in nodes:
            self.add_node(node)

    def rename_node(self, old, new):
        """
        >>> g = Tree()
        >>> g.add_nodes_from([0, 1, 2, 3, 4, 5, 6])
        >>> g.add_edges_from([(0, 1), (0, 2), (0, 3), (0, 6), (3, 4), (3, 5)])

        >>> oldObj = g[0].object
        >>> g.rename_node(0, 11)
        >>> g.nodes()
        [1, 2, 3, 4, 5, 6, 11]
        >>> g.edges()
        [(3, 4), (3, 5), (11, 1), (11, 2), (11, 3), (11, 6)]
        >>> g[11]
        <Vertex 11 (0)>
        >>> g[11].object == oldObj
        True
        """
        if old == new:
            return
        super(Tree, self).rename_node(old, new)
        vertexNew = self.vertexLookup.get(old)
        if not vertexNew:
            return
        vertexNew.id = new
        self.vertexLookup[new] = vertexNew
        del self.vertexLookup[old]

    def remove_node(self, nodeID):
        super(Tree, self).remove_node(nodeID)
        del self.vertexLookup[nodeID]

    def add_edge(self, parentID, childID, data=1):
        if parentID == childID:
            return
        super(Tree, self).add_edge(parentID, childID, data)

    def add_edges_from(self, edges):
        for edge in edges:
            if len(edge) == 2:
                fromNode, toNode = edge
                data = None
            elif len(edge) == 3:
                fromNode, toNode, data = edge
            self.add_edge(fromNode, toNode, data)

    def copy(self):
        H = super(Tree, self).copy()
        H.vertexLookup = self.vertexLookup
        return H



    def reindex(self, startingIndex=0):
        """
        >>> g = Tree()
        >>> g.add_nodes_from([0, 1, 2, 3, 4, 5, 6])
        >>> g.add_edges_from([(0, 1), (0, 2), (0, 3), (0, 6), (3, 4), (3, 5)])

        >>> g.reindex(2000)
        >>> g.nodes()
        [2000, 2001, 2002, 2003, 2004, 2005, 2006]
        >>> g.edges()
        [(2000, 2001), (2000, 2002), (2000, 2003), (2000, 2006), (2003, 2004), (2003, 2005)]
        >>> g.successors(2003)
        [2004, 2005]
        >>> g.successors(2000)
        [2001, 2002, 2003, 2006]

        >>> g.reindex(0)
        >>> g.nodes()
        [0, 1, 2, 3, 4, 5, 6]
        >>> g.edges()
        [(0, 1), (0, 2), (0, 3), (0, 6), (3, 4), (3, 5)]

        >>> g.reindex(3)
        >>> g.nodes()
        [3, 4, 5, 6, 7, 8, 9]
        >>> g.edges()
        [(3, 4), (3, 5), (3, 6), (3, 9), (6, 7), (6, 8)]
        """
        largestID = self.get_last_id()
        # so as not to encounter any id-overlap, shift the ids up if need be,
        # and then reindex again with the actual, desired starting id
        if startingIndex <= largestID:
            self.reindex(largestID + 1)
        for new, old in enumerate(sorted(self.nodes())):
            self.rename_node(old, new + startingIndex)

    def check_nodes(self):
        """
        >>> g = Tree()
        >>> g.add_nodes_from([0, 1, 2, 3, 4, 5, 6])
        >>> g.add_edges_from([(0, 1), (0, 2), (0, 3), (0, 6), (3, 4), (3, 5)])

        >>> g.leafNodes
        []
        >>> g.branchNodes
        []
        >>> g.check_nodes()
        >>> g.leafNodes
        [1, 2, 4, 5, 6]
        >>> g.branchNodes
        [0, 3]
        """
        for node in self.nodes():
            if self.successors(node):
                self.branchNodes.append(node)
            else:
                self.leafNodes.append(node)

    def get_branch_nodes(self, force_check=False):
        """
        >>> g = Tree()
        >>> g.add_nodes_from([0, 1, 2, 3, 4, 5, 6])
        >>> g.add_edges_from([(0, 1), (0, 2), (0, 3), (0, 6), (3, 4), (3, 5)])

        >>> g.get_branch_nodes()
        [0, 3]
        """
        if force_check or not self.branchNodes:
            self.check_nodes()
        return self.branchNodes

    def get_leaf_nodes(self, force_check=False):
        """
        >>> g = Tree()
        >>> g.add_nodes_from([0, 1, 2, 3, 4, 5, 6])
        >>> g.add_edges_from([(0, 1), (0, 2), (0, 3), (0, 6), (3, 4), (3, 5)])

        >>> g.get_leaf_nodes()
        [1, 2, 4, 5, 6]
        """
        if force_check or not self.leafNodes:
            self.check_nodes()
        return self.leafNodes

    def get_sub_nodes(self, nodeID):
        """
        For the given node id, find all children nodes and return node id + the
        list of the children nodes.

        >>> g = Tree()
        >>> g.add_nodes_from([0, 1, 2, 3, 4, 5, 6])
        >>> g.add_edges_from([(0, 1), (0, 2), (0, 3), (0, 6), (3, 4), (3, 5)])

        >>> g.get_sub_nodes(2)
        [2]
        >>> g.get_sub_nodes(3)
        [3, 4, 5]
        """
        nodes = []
        def descend(nodeID):
            nodes.append(nodeID)
            for child in self.successors(nodeID):
                nodes.append(child)
                if self.successors(child):
                    descend(child)
        descend(nodeID)
        return list(set(nodes))

    def get_last_id(self):
        """
        >>> g = Tree()
        >>> g.add_node(0)
        >>> g.get_last_id()
        0
        >>> g.add_node(1)
        >>> g.get_last_id()
        1
        >>> g.add_node(7)
        >>> g.get_last_id()
        7
        """
        return sorted(self.vertexLookup.keys())[-1]

    def get_subgraph(self, nodes, copy=True, reindex=False):
        g = super(Tree, self).subgraph(nodes, copy=copy)
        g.vertexLookup = {}
        for node in nodes:
            g.vertexLookup.update({node: self.vertexLookup[node]})
        if reindex:
            g.reindex()
        return g

    def add_subgraph(self, nodeID, subgraph):
        """
        Add to the graph without removing or displacing other graph nodes.

        Main graph:
        >>> g = Tree()
        >>> g.add_node(1)

        Create what will be the subgraph:
        >>> s1 = Tree()
        >>> s1.add_nodes_from([0, 1, 2, 3, 4, 5])
        >>> s1.add_edges_from([(0, 1), (0, 2), (1, 3), (3, 4), (3, 5)])

        >>> g.add_subgraph(1, s1)
        >>> g.nodes()
        [1, 2, 3, 4, 5, 6, 7]
        >>> g.edges()
        [(1, 2), (2, 3), (2, 4), (3, 5), (5, 6), (5, 7)]

        >>> g.add_subgraph(1, Graph())
        Traceback (most recent call last):
        ...
        GraphError: Subgraph must be of the same type
        """
        if not isinstance(subgraph, Tree):
            raise GraphError, "Subgraph must be of the same type"
        # get the last id number used in the graph, increment by 1, and reindex
        # the subgraph
        startingID = self.get_last_id() + 1
        subgraph.reindex(startingID)
        # add the newly indexed nodes to the main graph
        self.add_nodes_from(subgraph.nodes())
        # add the edges from the reindexed subgraph to the main graph
        self.add_edges_from(subgraph.edges())
        # add an edge from nodeID to the parent node of the subgraph
        self.add_edge(nodeID, subgraph[startingID].id)

    def remove_subgraph(self, nodeID):
        """
        Delete the given node and all of its children.

        >>> g = Tree()
        >>> g.add_nodes_from([0, 1, 2, 3, 4, 5, 6])
        >>> g.add_edges_from([(0, 1), (0, 2), (0, 3), (0, 6), (3, 4), (3, 5)])

        >>> g.remove_subgraph(2)
        Traceback (most recent call last):
        ...
        GraphError: The given node is not the root of a subgraph

        >>> g.remove_subgraph(3)
        >>> g.nodes()
        [0, 1, 2, 6]
        >>> g.edges()
        [(0, 1), (0, 2), (0, 6)]
        """
        if nodeID not in self.get_branch_nodes():
            raise GraphError, "The given node is not the root of a subgraph"
        for node in self.get_sub_nodes(nodeID):
            self.remove_node(node)

    def replace_subgraph(self, nodeID, subgraph):
        """
        Replace the given node and its children with the given subgraph.

        Main graph:
        >>> g = Tree()
        >>> g.add_node(1)

        Create what will be the subgraph:
        >>> s1 = Tree()
        >>> s1.add_nodes_from([0, 1, 2, 3, 4, 5])
        >>> s1.add_edges_from([(0, 1), (0, 2), (1, 3), (3, 4), (3, 5)])
        >>> g.add_subgraph(1, s1)

        The subgraph was connected to node 1; the subgraph's top node in the
        larger graph will have gotten reindexed from 0 to 2.

        Create the subgraph that will replace the first subgraph:
        >>> s2 = Tree()
        >>> s2.add_nodes_from([0, 1, 2])
        >>> s2.add_edges_from([(0, 1), (0, 2)])

        Replace the old subgraph with the new one:
        >>> g.replace_subgraph(2, s2)

        Check the results
        >>> g.nodes()
        [1, 2, 3, 4]
        >>> g.edges()
        [(1, 2), (2, 3), (2, 4)]
        """
        # we can get away with using the first index because in a tree, there
        # should only ever be one parent for any given node
        original_parent = self.predecessors(nodeID)[0]
        op = original_parent
        self.remove_subgraph(nodeID)
        self.add_subgraph(nodeID, subgraph)
        # when re deleted the subgraph, we also removed its connection to the
        # parent, so we need to re-add that
        self.add_edge(original_parent, nodeID)


if __name__ == '__main__':
    import doctest
    doctest.testmod()
