#-*- coding: utf-8 -*-
"""
A trivial Lisp implementation.
"""
import sys

from pylispng import parser
from pylispng import symbol
from pylispng.util import property_cache
from pylispng.exception import Error, UnboundSymbolError


class Environment(object):

    def __init__(self, parent=None, bindings=None):
        if bindings:
            self.e = bindings
        else:
            self.e = {}
        # Parent environment.
        self.parent = parent
        if self.parent:
            self.level = self.parent.level + 1
        else:
            self.level = 0
        self.init()

    def __repr__(self):
        s = "\nLevel %s:\n" % self.level
        keys = self.e.keys()
        keys.sort()
        for key in keys:
            if key != "*env*":
                s = s + " %10s: %s\n" % (key, self.e[key])
        if self.parent:
            return s + `self.parent`# + s
        else:
            return s

    def init(self):
        """
        Subclass customization code here.
        """
        pass

    def set(self, sym, value):
        """
        Assign a binding in the environment.

        Keeps looking for the symbol in outer environment until found or
        until it gets to the top-level environment.
        """
        if self.e.has_key(sym):
            self.e[sym] = value
        elif self.parent:
            self.parent.set(sym, value)
        else:
            self.e[sym] = value

    def lexicalSet(self, sym, value):
        """
        Set a binding in the immediate environment.
        """
        #print "LEXICAL_SET(%i):" % self.level, sym, value
        self.e[sym] = value

    def get(self, sym):
        if not self.e.has_key(sym):
            # Consult the parent environment if the symbol isn't here.
            if self.parent:
                return self.parent.get(sym)
            else:
                raise UnboundSymbolError, sym
        else:
            #print "GET(%i)" % self.level, sym, self.e[sym]
            return self.e[sym]

    def pop(self):
        """
        Throw myself away, returning the parent environment.
        """
        return self.parent

    def push(self, bindings=None):
        """
        Create a new environment.
        """
        return Environment(self, bindings=bindings)

    def nullp(self):
        return symbol.FALSE


class Interpreter(object):
    """
    A small lisp environment.
    """
    envData = {
        # The default tiny lisp environment.
        '*self*': ['self'],
        '*env*': ['self.e'],
        '*true*': [symbol.TRUE],
        '*false*': [symbol.FALSE],
        '*true-enough*': [symbol.TRUE],
        '*gensym-string*': [symbol.SymbolObject, 'GENSYM@', 'no_self'],
        '*gensym-counter*': [symbol.NumberObject, 0, 'no_self'],
        'print': [symbol.FunctionObject, 'do_print'],
        'first': [symbol.FunctionObject, 'do_first'],
        'car': [symbol.FunctionObject, 'do_first'],
        'replaca': [symbol.FunctionObject, 'do_replaca'],
        'second': [symbol.FunctionObject, 'do_second'],
        'cadr': [symbol.FunctionObject, 'do_second'],
        'third': [symbol.FunctionObject, 'do_third'],
        'rest': [symbol.FunctionObject, 'do_rest'],
        'cdr': [symbol.FunctionObject, 'do_rest'],
        'replacd': [symbol.FunctionObject, 'do_replacd'],
        'cons': [symbol.FunctionObject, 'do_cons'],
        'append': [symbol.FunctionObject, 'do_append'],
        'null?': [symbol.FunctionObject, 'do_nullp'],
        'list': [symbol.FunctionObject, 'do_list'],
        'assoc': [symbol.FunctionObject, 'do_assoc'],
        '+': [symbol.FunctionObject, 'do_add'],
        '-': [symbol.FunctionObject, 'do_sub'],
        '/': [symbol.FunctionObject, 'do_div'],
        '%': [symbol.FunctionObject, 'do_mod'],
        '*': [symbol.FunctionObject, 'do_mul'],
        '>': [symbol.FunctionObject, 'do_gt'],
        '>=': [symbol.FunctionObject, 'do_ge'],
        '<': [symbol.FunctionObject, 'do_lt'],
        '<=': [symbol.FunctionObject, 'do_le'],
        '==': [symbol.FunctionObject, 'do_eql'],
        'eql': [symbol.FunctionObject, 'do_eql'],
        '!=': [symbol.FunctionObject, 'do_neq'],
        'not': [symbol.FunctionObject, 'do_not'],
        'and': [symbol.FunctionObject, 'do_and'],
        'or': [symbol.FunctionObject, 'do_or'],
        'logic': [symbol.FunctionObject, 'do_logic'],
        'load': [symbol.FunctionObject, 'do_read'],
        'gensym': [symbol.FunctionObject, 'do_gensym'],
        'eval': [symbol.FunctionObject, 'do_eval'],
        'list?': [symbol.FunctionObject, 'do_listp'],
        'pair?': [symbol.FunctionObject, 'do_pairp'],
        'symbol?': [symbol.FunctionObject, 'do_symbolp'],
        'string?': [symbol.FunctionObject, 'do_stringp'],
        'number?': [symbol.FunctionObject, 'do_numberp'],
        'logic?': [symbol.FunctionObject, 'do_logicp'],
        'symbol-name': [symbol.FunctionObject, 'do_symbol_name'],
        'top-level': [symbol.FunctionObject, 'do_top_level'],
        'the-environment': [symbol.FunctionObject, 'do_the_environment'],
        'get-environment': [symbol.FunctionObject, 'do_get_environment'],
        'env-get': [symbol.FunctionObject, 'do_env_get'],
        'env-set': [symbol.FunctionObject, 'do_env_set'],
        'get': [symbol.FunctionObject, 'do_get'],
        'put': [symbol.FunctionObject, 'do_put'],
        'elt': [symbol.FunctionObject, 'do_elt'],
        'apply': [symbol.FunctionObject, 'do_apply'],
        'send': [symbol.FunctionObject, 'do_send'],
        # Chat with Python.
        #'py-setattr': [symbol.FunctionObject, 'do_py_setattr'],
        #'py-getattr': [symbol.FunctionObject, 'do_py_getattr'],
        #'py-hash-make': [symbol.FunctionObject, 'do_py_hash_make'],
        #'py-hash-get': [symbol.FunctionObject, 'do_py_hash_get'],
        #'py-hash-set': [symbol.FunctionObject, 'do_py_hash_set'],
        #'py-send': [symbol.FunctionObject, 'do_py_send'],
        'py-format': [symbol.FunctionObject, 'do_py_format'],
        'py-type': [symbol.FunctionObject, 'do_py_type'],
        'py-eval': [symbol.FunctionObject, 'do_py_eval'],
        'py-exec': [symbol.FunctionObject, 'do_py_eval'],
        'keyword->pyhash': [symbol.FunctionObject, 'do_kw_to_pyhash'],
        # macros/syntax
        'begin': [symbol.SyntaxObject, 'do_m_begin'],
        'setq': [symbol.SyntaxObject, 'do_m_setq'],
        'def': [symbol.SyntaxObject, 'do_m_setq'],
        'let': [symbol.SyntaxObject, 'do_m_let'],
        'lambda': [symbol.SyntaxObject, 'do_m_lambda'],
        'macro': [symbol.SyntaxObject, 'do_m_macro'],
        'macro-expand': [symbol.SyntaxObject, 'do_m_macro_expand'],
        'quote': [symbol.SyntaxObject, 'do_m_quote'],
        'i-quote': [symbol.SyntaxObject, 'do_m_iquote'],
        'if': [symbol.SyntaxObject, 'do_m_if'],
        'cond': [symbol.SyntaxObject, 'do_m_cond'],
        'with-py-hash-env': [symbol.SyntaxObject, 'do_m_with_hash'],
        }

    def __init__(self, iostreams=(sys.stdin, sys.stdout, sys.stderr)):
        (self.stdin, self.stdout, self.stderr) = iostreams
        self.r = parser.Parser()
        self.e = Environment()
        for symbol, args in self.envData.items():
            self.prepareEnv(symbol, args)
        # Subclasses should override init() to add new functions
        # and syntax.
        self.init()

    def init(self):
        """
        Secondary initialization

        Subclasses should add new functionality to PyLisp in this method.
        """
        pass

    def addPoundHelper(self, char, helper):
        self.r.add_pound_helper(char, helper)

    def prepareEnv(self, sym, args):
        if len(args) == 1:
            if args[0] == 'self':
                arg = self
            elif args[0] == 'self.e':
                arg = self.e
            else:
                arg = args[0]
        elif args[-1] == 'no_self':
            klass, param, dummy = args
            arg = klass(param)
        else:
            klass, attr = args
            arg = klass(getattr(self, attr))
        self.intern(sym, arg)

    def getSymbolType(sym):
        """

        """
        if Interpreter.envData.has_key(sym):
            type = Interpreter.envData[sym][0]
        else:
            type = symbol.SymbolObject
        return type

    getSymbolType = staticmethod(getSymbolType)

    def intern(self, sym, val):
        """
        assign a symbol value in the environment
        """
        self.e.set(sym, val)

    def lexicalIntern(self, sym, val):
        """
        assign a symbol value in the immediate environment
        """
        self.e.lexicalSet(sym, val)

    def lookup(self, sym):
        return self.e.get(sym, UnboundSymbolError)

    def set(self, sym, propery, value):
        if not property_cache.has_key(sym):
            property_cache[sym] = {}
        property_cache[sym][property] = value

    def get(self, sym, property):
        try:
            return property_cache[sym][property]
        except:
            return None

    def eval(self, form):
        """
        Evaluate a lisp form, returning the value.
        """
        return form.eval(self.e)

    def evalString(self, str):
        return self.eval(self.r.getSExpr(str))

    def rawInput(self, prompt) :
        """
        rawInput with pseudo I/O streams.
        """
        if prompt :
            self.stdout.write("%s" % prompt)
            self.stdout.flush()
        line = self.stdin.readline()
        if line[-1] == "\n":
            line = line[:-1]
        return line

    # Code from Brian P Templeton <bpt@tunes.org>.  Now you can enter
    # multi-line sexps from the interactive interface.  *Much* nicer.
    def readFullSExp(self, line="", parenlevel=0):
        """
        get a complete sexp
        """
        if line != "":
            line = line + " "
        if self.e.level != 0:
            prompt = "lisp %i%s " % (self.e.level, ">" * (parenlevel+1))
        else:
            prompt = "lisp%s " % (">" * (parenlevel+1))
            line = line  + self.rawInput(prompt)
            oparens = 0
            cparens = 0
            for c in line:
                if c == "(":
                    oparens = oparens + 1
                elif c == ")":
                    cparens = cparens + 1
            if oparens > cparens:
                return self.readFullSExp(line, parenlevel+1)
            else:
                return line

    def read(self, file):
        """
        l.read(file) - read in and evaluate a file of PyLisp code
        """
        f = open(file, "r")
        code = f.readlines()
        f.close()
        code = re.sub(";.*(\n|$)", "", string.join(code))
        sexpr = self.r.getSExpr(code)
        while sexpr:
            self.eval(sexpr)
            sexpr = self.r.getSExpr()

    def repl(self):
        """
        Interactive lisp loop.

        A much friendlier version of the read-eval-print loop, care
        of Ole Martin Bjrndalen.
        """
        while 1:
            try:
                line = self.readFullSExp()
            except EOFError:
                self.stdout.write("\n")
                break
            if not line:
                continue
            if line in ['(quit)', '(bye)']:
                break
            try:
                sexpr = self.r.getSExpr(line)
                while sexpr:
                    self.stdout.write("\t%s\n" % self.eval(sexpr))
                    sexpr = self.r.getSExpr()
            except UnboundSymbolError, e:
                self.stderr.write("ERROR: unbound symbol '%s'\n" % e.args[0])
            except:
                import traceback, sys
                traceback.print_exc(self.stderr)

    def replD(self):
        """
        l.replD() - debugging read-eval-print loop
        """
        # allows access to the Python debugger when something dies
        line = self.rawInput("lisp> ")
        while line and line not in ['(quit)', '(bye)']:
            sexpr = self.r.getSExpr(line)
            while sexpr:
                self.stdout.write("\t %s\n" % self.eval(sexpr))
                sexpr = self.r.getSExpr()
            line = self.rawInput("lisp> ")

    def pushE(self, env=None):
        if env:
            self.e = self.e.push(env)
        else:
            self.e = self.e.push()

    def popE(self):
        self.e = self.e.pop()

    def unwindE(self):
        """
        Unroll to the top environment level.
        """
        while self.e.parent:
            self.e = self.e.parent

    def do_print(self, env, args):
        for a in args:
            result = a.eval(env)
            self.stdout.write("%s " % str(result))
        self.stdout.write("\n")
        return symbol.TRUE

    def do_first(self, env, args):
        return args[0].eval(env).first()

    def do_replaca(self, env, args):
        return args[0].eval(env).replaca(args[1].eval(env))

    def do_second(self, env, args):
        return args[0].eval(env).second()

    def do_third(self, env, args):
        return args[0].eval(env).third()

    def do_rest(self, env, args):
        return args[0].eval(env).rest()

    def do_replacd(self, env, args):
        return args[0].eval(env).replacd(args[1].eval(env))

    def do_cons(self, env, args):
        return args[1].eval(env).cons(args[0].eval(env))

    def do_append(self, env, args):
        lst = args[0].eval(env)
        for l in args[1:]:
            lst.append(l.eval(env))
        return lst

    def do_nullp(self, env, args):
        if len(args[0].eval(env)) == 0:
            return symbol.TRUE
        else:
            return symbol.FALSE

    def do_list(self, env, args):
        ans = []
        for n in args:
            ans.append(n.eval(env))
        return symbol.ListObject(ans)

    def do_assoc(self, env, args):
        return args[1].eval(env).assoc(args[0].eval(env))

    def do_add(self, env, args):
        ans = args[0].eval(env)
        for n in args[1:]:
            ans = ans + n.eval(env)
        return symbol.NumberObject(ans)

    def do_sub(self, env, args):
        ans = args[0].eval(env)
        if args[1:]:
            for n in args[1:]:
                ans = ans - n.eval(env)
        else:
            # (- 5) should negate
            ans = -ans.value
        return symbol.NumberObject(ans)

    def do_div(self, env, args):
        ans = args[0].eval(env)
        if args[1:]:
            for n in args[1:]:
                ans = ans / n.eval(env)
        else:
            # (/ 2) - give inverse
            ans = 1 / ans.value
        return symbol.NumberObject(ans)

    def do_mod(self, env, args):
        ans = args[0].eval(env)
        for n in args[1:]:
            ans = ans % n.eval(env)
        return symbol.NumberObject(ans)

    def do_mul(self, env, args):
        ans = args[0].eval(env)
        for n in args[1:]:
            ans = ans * n.eval(env)
        return symbol.NumberObject(ans)

    def do_gt(self, env, args):
        if args[0].eval(env) > args[1].eval(env):
            return symbol.TRUE
        else:
            return symbol.FALSE

    def do_ge(self, env, args):
        if args[0].eval(env) >= args[1].eval(env):
            return symbol.TRUE
        else:
            return symbol.FALSE


    def do_lt(self, env, args):
        if args[0].eval(env) < args[1].eval(env):
            return symbol.TRUE
        else:
            return symbol.FALSE

    def do_le(self, env, args):
        if args[0].eval(env) <= args[1].eval(env):
            return symbol.TRUE
        else:
            return symbol.FALSE

    def do_eql(self, env, args):
        a = args[0].eval(env)
        b = args[1].eval(env)
        if (isinstance(a, symbol.SymbolObject) and
            isinstance(b, symbol.SymbolObject)):
            if a.name == b.name:
                return symbol.TRUE
            else:
                return symbol.FALSE
        else:
            if a == b:
                return symbol.TRUE
            else:
                return symbol.FALSE

    def do_neq(self, env, args):
        if args[0].eval(env) != args[1].eval(env):
            return symbol.TRUE
        else:
            return symbol.FALSE

    def do_not(self, env, args):
        return -args[0].eval(env)

    def do_and(self, env, args):
        ans = args[0].eval(env)
        for arg in args[1:]:
            ans = ans & arg.eval(env)
            # Short circuit once falsity is known.
            if ans.value == 0.0:
            #if ans.nullp():
                break
        return ans

    def do_or(self, env, args):
        truth = env.get('*true-enough*')
        ans = args[0].eval(env)
        for arg in args[1:]:
            ans = ans | arg.eval(env)
            # Short circuit once truth is known.
            if ans.value >= truth:
                break
        return ans

    # All logic in PyLisp is fuzzy.  If you only use *true* and *false*
    # then you get normal T/F logic.  Once you introduce a fuzzy logic
    # value -- such as (logic 0.34) -- then the fuzziness propagates
    # as appropriate.  The normal lisp notion that anything that isn't
    # nil is true does *not* apply to this lisp.
    def do_logic(self, env, args):
        return symbol.LogicObject(args[0].eval(env))

    def do_read(self, env, args):
        self.read(args[0].eval(env))

    # Useful for complex macros.
    def do_gensym(self, env, args):
        env.set('*gensym-counter*', env.get('*gensym-counter*') + 1)
        return symbol.SymbolObject("%s%s" % (env.get('*gensym-string*'),
                                      env.get('*gensym-counter*')))

    def do_eval(self, env, args):
        return args[0].eval(env).eval(env)

    def do_listp(self, env, args):
        if isinstance(args[0].eval(env), symbol.ListObject):
            return symbol.TRUE
        else:
            return symbol.FALSE

    def do_pairp(self, env, args):
        if isinstance(args[0].eval(env), symbol.PairObject):
            return symbol.TRUE
        else:
            return symbol.FALSE

    def do_stringp(self, env, args):
        if isinstance(args[0].eval(env), symbol.StringObject):
            return symbol.TRUE
        else:
            return symbol.FALSE

    def do_symbolp(self, env, args):
        if isinstance(args[0].eval(env), symbol.SymbolObject):
            return symbol.TRUE
        else:
            return symbol.FALSE

    def do_numberp(self, env, args):
        if isinstance(args[0].eval(env), symbol.NumberObject):
            return symbol.TRUE
        else:
            return symbol.FALSE

    def do_logicp(self, env, args):
        if isinstance(args[0].eval(env), symbol.LogicObject):
            return symbol.TRUE
        else:
            return symbol.FALSE

    def do_symbol_name(self, env, args):
        sym = args[0].eval(env)
        return symbol.StringObject(sym.name)

    def do_top_level(self, env, args):
        self.unwindE()
        return symbol.TRUE

    def do_the_environment(self, env, args):
        return self.e

    def do_get_environment(self, env, args):
        # Yank the environment bundled up with a lambda or macro
        form = args[0].eval(env)
        if (isinstance(form, symbol.LambdaObject) or
            isinstance(form, symbol.MacroObject)):
            return form.env
        else:
            raise Error, "get-environment: must take lambda or macro"

    def do_env_get(self, env, args):
        envform = args[0].eval(env)
        if isinstance(envform, Environment):
            return envform.get(args[1].eval(env).name)
        else:
            raise Error, "env-get: must get environment as first argument"

    def do_env_set(self, env, args):
        envform = args[0].eval(env)
        if isinstance(envform, Environment):
            value = args[2].eval(env)
            envform.set(args[1].eval(env).name, value)
            return value
        else:
            raise Error, "env-set: must get environment as first argument"

    def do_put(self, env, args):
        val = args[2].eval(env)
        args[0].eval(env).put(args[1].eval(env).name, val)
        return val

    def do_get(self, env, args):
        return args[0].eval(env).get(args[1].eval(env).name)

    def do_elt(self, env, args):
        return args[1].eval(env)[args[0].eval(env).value]

    def do_apply(self, env, args):
        # This shuffles the args about a bit, conses the function, then evals.
        return args[1].eval(env).cons(args[0]).eval(env)

    def do_send(self, env, args):
        return None

    def do_py_format(self, env, args):
        eargs = []
        for arg in args[1:]:
            eargs.append(arg.eval(env))
        fmt = args[0].eval(env)
        if isinstance(fmt, symbol.StringObject):
            return fmt.string % tuple(eargs)
        elif type(fmt) == types.StringType:
            return fmt % tuple(eargs)
        else:
            raise Error, "py-format: format must be a string"

    def do_py_type(self, env, args):
        """
        (py-type o) - try to convert object o into its Python value
        """
        # This is most useful when trying to extend PyLisp with external
        # Python capabilities.
        o = args[0].eval(env)
        if (isinstance(o, symbol.NumberObject) or
            isinstance(o, symbol.LogicObject)):
            return o.value
        elif isinstance(o, symbol.StringObject):
            return o.string
        elif isinstance(o, symbol.ListObject):
            return o.list
        else:
            return o

    def do_py_eval(self, env, args):
        """
        (py-eval "python code") - evaluate a python expression
        """
        return eval(args[0].eval(env))

    def do_py_exec(self, env, args):
        """
        (py-exec "python code") - execute python code
        """
        exec(args[0].eval(env))

    # MOVE THIS TO class ListObject!!!!
    def do_kw_to_pyhash(self, env, args):
        """
        (keyword->pyhash '(a b :text "this is cool" :rate 0.53))

        This turns Lisp-style keyword lists into a python dictionary.
        """
        hsh = {}
        margs = args[0].eval(env).list
        n = len(margs)
        i = 0
        while i < n:
            sym = margs[i]
            if isinstance(sym, symbol.SymbolObject) and sym.name[0] == ':':
                hsh[sym.name[1:]] = margs[i + 1]
                i = i + 1
            i = i + 1
        return hsh

    # macros/syntax
    def do_m_begin(self, env, args):
        answer = symbol.FALSE
        for code in args:
            answer = code.eval(env)
        return answer

    def do_m_setq(self, env, args):
        self.intern(args[0].name, args[1].eval(env))
        return self.e.get(args[0].name)

    def do_m_let(self, env, args):
        """
        Transform the let into a lambda
        (let ((x 15) (y (+ 2 3)) z) code) ==
        ((lambda (x y z) code) 15 (+ 2 3) '())
                 ^^^^^^^       ^^ ^^^^^^^ ^^^
                 bindings         vallist
        """
        bindings = args[0]
        code = args[1:]
        arglist = []
        vallist = []
        # Rip out the binding names and their initializers if they exist.
        for b in bindings:
            if isinstance(b, symbol.SymbolObject):
                # bare symbol - no initializer
                arglist.append(b)
                vallist.append(
                    symbol.ListObject([symbol.QUOTE, symbol.ListObject([])]))
            elif isinstance(b, symbol.ListObject):
                # list form - symbol and initializer
                arglist.append(b.first())
                vallist.append(b.second())
            else:
                raise Error, "malformed (let ...)"
        # Pull the whole mess together into the lambda form, then eval.
        return symbol.ListObject([
            symbol.ListObject([
                symbol.SymbolObject('lambda'),
                symbol.ListObject(arglist)] + code)] + vallist).eval(env)

    def do_m_lambda(self, env, args):
        # (lambda (x y z) (* x y z))
        if self.e == env.get('*env*'):
            #print "NULL DEFINED ENV"
            return symbol.LambdaObject(args[0], args[1:], None)
        else:
            #print "LEXICALLY DEFINED ENV"
            return symbol.LambdaObject(args[0], args[1:], env)

    def do_m_macro(self, env, args):
        # (def incq (macro (x) `(setq ,x (+ ,x 1)))
        if self.e == env.get('*env*'):
            #print "NULL DEFINED ENV"
            return symbol.MacroObject(args[0], args[1:], None)
        else:
            #print "LEXICALLY DEFINED ENV"
            return symbol.MacroObject(args[0], args[1:], env)

    def do_m_macro_expand(self, env, args):
        """
        Grab the first/only argument, pull out the first element of
        that list and eval it -- it is a symbol which had better
        resolve to a macro -- expand with the rest of its arguments,
        which have to be ripped out of the argument list to trick
        the expand method into looking like an eval call.  Return
        the result, which is a list of PyLisp code.
        """
        return args[0].first().eval(env).expand(env, args[0].rest())

    def do_m_quote(self, env, args):
        return args[0]

    def do_m_iquote(self, env, args):
        if isinstance(args[0], symbol.ListObject):
            return args[0].iquote(env)
        else:
            return args[0]

    def do_m_if(self, env, args):
        bool = args[0].eval(env)
        if isinstance(bool, symbol.LogicObject):
            if bool >= env.get('*true-enough*'):
                return args[1].eval(env)
            else:
                if len(args) == 3:
                    return args[2].eval(env)
                else:
                    return bool
        elif hasattr(bool, 'nullp'):
            # Null is false.
            #print "hasattr nullp"
            if args[0].eval(env).nullp() == symbol.TRUE:
                if len(args) == 3:
                    return args[2].eval(env)
                else:
                    return bool
            else:
                return args[1].eval(env)
        else:
            # This is also testing for the false case.
            if bool in [0, 0.0, ""]:
                if len(args) == 3:
                    return args[2].eval(env)
                else:
                    return bool
            else:
                return args[1].eval(env)

    def do_m_cond(self, env, args):
        """
        (cond (test0 return0)
              (test1 return1)
              (*true* returnt) )
        """
        for clause in args:
            bool = clause.first().eval(env)
            if isinstance(bool, symbol.LogicObject):
                if bool >= env.get('*true-enough*'):
                    return clause.second().eval(env)
            elif hasattr(bool, 'nullp'):
                # Null is false.
                #print "hasattr nullp"
                if bool.nullp() == symbol.TRUE:
                    pass
                else:
                    return clause.second().eval(env)
            else:
                # This is also testing for the false case.
                if bool in [0, 0.0, ""]:
                    pass
                else:
                    return clause.second().eval(env)
        return symbol.FALSE

    def do_m_with_hash(self, env, args):
        """
        (with-py-hash-env some-hash-table ...)
        """
        hsh = args[0].eval(env)
        # Set up the new environment.
        self.pushE()
        # Intern the new values.
        for (key, value) in hsh.items():
            if type(value) == types.StringType:
                self.intern(key, value)
            else:
                self.intern(key, self.r.getSExpr(`value`))
        # Eval the code.
        for code in args[1:]:
            answer = code.eval(env)
        # Pop off the hash-table environemt.
        self.popE()
        # Return final form value?  Perhaps this should return the final
        # environment somehow, too?  Or require this:
        #   (with-py-has-env some-hash
        #     ...
        #     ...
        #     (the-environment))
        return answer

def getObj(sym):
    """
    A utility function that parses a simple token and returns an appropriate
    PyLisp object representation of that token.
    """
    return parser.Parser.get_simple_token(str(sym))

def getType(sym):
    """
    A utility function that performs a symbol look-up and returns a lisp
    environment object for that symbol.
    """
    if isinstance(sym, symbol.ListObject):
        return symbol.ListObject
    return Interpreter.getSymbolType(str(sym))
