"""
A trivial Lisp implementation.
"""
import re
import string

from pyparsing import ParseResults

from pylispng import grammar
from pylispng import symbol
from pylispng.exception import ParseError


SPECIAL = "()`',@"
SYNTAX = string.whitespace + SPECIAL


class Parser(object):
    """
    >>> p = Parser()
    >>> len(p)
    0
    >>> p.str
    ''
    >>> p.strlen
    0
    >>> p.sexpr
    []
    >>> p.tokenize()
    >>> p.getSExpr()

    >>> p = Parser('(+ 1 2)')
    >>> len(p)
    3
    >>> p.str
    '(+ 1 2)'
    >>> p.strlen
    7
    >>> p.sexpr
    (+ 1 2)
    >>> type(p.sexpr)
    <class 'pylispng.symbol.ListObject'>

    >>> p = Parser('(+ 1 2 (* 5 3 (* (- 4 7) (* 14 3))) 4)')
    >>> len(p)
    14
    >>> p.str
    '(+ 1 2 (* 5 3 (* (- 4 7) (* 14 3))) 4)'
    >>> p.strlen
    38
    >>> p.sexpr
    (+ 1 2 (* 5 3 (* (- 4 7) (* 14 3))) 4)
    >>> p.tokens
    ['+', '1', '2', '*', '5', '3', '*', '-', '4', '7', '*', '14', '3', '4']
    >>> p.symbols == list(p)
    True
    >>> p.symbols
    [+, <number 1>, <number 2>, *, <number 5>, <number 3>, *, -, <number 4>, <number 7>, *, <number 14>, <number 3>, <number 4>]
    >>> [str(x) for x in p]
    ['+', '1', '2', '*', '5', '3', '*', '-', '4', '7', '*', '14', '3', '4']
    >>> [str(x) for x in list(p)]
    ['+', '1', '2', '*', '5', '3', '*', '-', '4', '7', '*', '14', '3', '4']
    >>> p[0]
    +
    >>> p[5].value
    3
    >>> p[len(p) - 3].value
    14
    >>> p[len(p) - 1].value
    4
    """
    def __init__(self, str=""):
        self.str = str
        self.i = 0
        self.pounds = {}
        self.tokens = []
        self.symbols = []
        if not str:
            self.strlen = 0
            self.sexpr = []
        else:
            self.strlen = len(str)
            self.sexpr = self.getSExpr()

    def __len__(self):
        return len(self.descend(self.sexpr))

    def __iter__(self):
        return iter(self.descend(self.sexpr))

    def __getitem__(self, itemIndex):
        for index, item in enumerate(self):
            if index == itemIndex:
                return item

    def descend(self, iterable):
        items = []
        if isinstance(iterable, symbol.SymbolObject):
            items.append(iterable)
            return items
        for item in iterable:
            if isinstance(item, symbol.ListObject):
                items.extend(self.descend(item))
            else:
                items.append(item)
        return items

    def addPoundHelper(self, char, helper):
        self.pounds[char] = helper

    def tokenize(self):
        if self.i >= self.strlen:
            return None
        # Munch leading whitespace.
        while self.i < self.strlen and self.str[self.i] in string.whitespace:
            self.i = self.i + 1
        if self.i == self.strlen:
            return None
        # Now, tokenize.
        if self.str[self.i] == '#':
            # Look ahead to get the second character of the pound escape
            # then sling on the next token for special treatment.
            self.i = self.i + 2
            return self.pounds[self.str[self.i - 1]](self.tokenize())
        if self.str[self.i] in SPECIAL:
            self.i = self.i + 1
            return self.str[self.i - 1]
        elif self.str[self.i] == '"':
            # Parse a string.
            str = ""
            self.i = self.i + 1
            while self.str[self.i] != '"' and self.i < self.strlen:
                if self.str[self.i] == '\\':
                    self.i = self.i + 1
                    spchar = self.str[self.i]
                    if spchar == "n":
                        str = str + "\n"
                    elif spchar == "t":
                        str = str + "\t"
                else:
                    str = str + self.str[self.i]
                self.i = self.i + 1
            # Remove trailing quote
            self.i = self.i + 1
            return symbol.StringObject(str)
        else:
            tok = ""
            # First, build the token.
            while self.i < self.strlen - 1:
                if self.str[self.i] in SYNTAX:
                    break
                else:
                    tok = tok + self.str[self.i]
                    self.i = self.i + 1
            if not self.str[self.i] in SYNTAX:
                tok = tok + self.str[self.i]
                self.i = self.i + 1
            # If the thing is a number, convert it; if it's a string, make it a
            # symbol
            return self.getSymbol(tok)

    def getSymbol(self, tok):
        """
        >>> p = Parser()
        >>> s = p.getSymbol('0')
        >>> type(s)
        <class 'pylispng.symbol.NumberObject'>
        >>> s
        <number 0>
        >>> s = p.getSymbol('1')
        >>> type(s)
        <class 'pylispng.symbol.NumberObject'>
        >>> s
        <number 1>
        >>> s = p.getSymbol('a')
        >>> type(s)
        <class 'pylispng.symbol.SymbolObject'>
        >>> s
        a
        >>> s = p.getSymbol('+')
        >>> type(s)
        <class 'pylispng.symbol.SymbolObject'>
        >>> s
        +
        """
        self.tokens.append(tok)
        try:
            sym = symbol.NumberObject(tok)
        except ValueError:
            sym = symbol.SymbolObject(tok)
        self.symbols.append(sym)
        return sym

    def getSExpr(self, str=None):
        """
        >>> p = Parser()
        >>> s = p.getSExpr('(+ 1 2)')
        >>> type(s)
        <class 'pylispng.symbol.ListObject'>
        >>> s
        (+ 1 2)

        >>> p.getSExpr('(+ 1 2 (* 3 4))')
        (+ 1 2 (* 3 4))
        """
        if str:
            self.i = 0
            self.str = str
            self.strlen = len(self.str)
        expr = None
        tok = self.tokenize()
        if tok == ')':
            raise ParseError, "Unexpected ')'"
        elif tok == "(":
            expr = []
            tok = self.tokenize()
            while tok != ")":
                if tok == '(':
                    # Start parsing again.
                    self.i = self.i - 1
                    expr.append(self.getSExpr())
                elif tok == "'":
                    expr.append(symbol.ListObject([
                        symbol.QUOTE, self.getSExpr()]))
                elif tok == "`":
                    expr.append(symbol.ListObject([
                        symbol.IQUOTE, self.getSExpr()]))
                elif tok == None:
                    raise ParseError, "Unexpected end of expression"
                else:
                    expr.append(tok)
                tok = self.tokenize()
            return symbol.ListObject(expr)
        elif tok == "'":
            tok = symbol.ListObject([symbol.QUOTE, self.getSExpr()])
        elif tok == "`":
            tok = symbol.ListObject([symbol.IQUOTE, self.getSExpr()])
        return tok


class NewParser(object):

    def __init__(self, string=""):
        self.grammar = grammar.getSExp()
        self.grammar.setParseAction(self._getSymbol)
        self.string = string
        self.sexpr = None
        if self.string:
            self.setSExpr(self.string)

    def _getSymbol(self, string, location, tokens):
        [token] = tokens
        #import pdb;pdb.set_trace()
        newToken = symbol.mapper.get(token)
        if not newToken:
            # We're going to skip pyparsing ParseResults objects for now
            if isinstance(token, ParseResults):
                return
            try:
                if re.match("^[0-9]", token):
                    newToken = symbol.NumberObject
            except Exception, error:
                import pdb;pdb.set_trace()
        if newToken:
            return [newToken(token)]

    def getLength(self):
        if self.sexpr is None:
            return 0
        return len(self.getSExpr())

    def getStringCount(self):
        return len(self.string)

    def setSExpr(self, string=""):
        self.sexpr = self.grammar.parseString(string)

    def getSExpr(self):
        if self.sexpr is not None:
            return self.sexpr.asList()[0]
        return None


if __name__ == "__main__":
    import doctest
    doctest.testmod()
