from pylispng import lisp
from pylispng import util
from pylispng import graph
from pylispng import parser
from pylispng import symbol
from pylispng.exception import GraphError, GraphIndexError


class ExpressionDescender(object):
    """
    This is a base class for any class that has the need to recursively descend
    a pyLisp-NG s-expression.
    """
    def handleOperator(self, operator, *args, **kwds):
        return operator

    def handleSymbol(self, symbol, *args, **kwds):
        return symbol

    def handleExpression(self, expression, *args, **kwds):
        return self.descend(expression)

    def descend(self, sexpr, *args, **kwds):
        """
        Recursively descend an s-expression.
        """
        output = []
        for node in sexpr:
            type = lisp.getType(node)
            if type == symbol.ListObject:
                output.extend(self.handleExpression(node, *args, **kwds))
            elif type == symbol.SymbolObject:
                output.append(self.handleSymbol(node, *args, **kwds))
            else:
                output.append(self.handleOperator(node, *args, **kwds))
        return output


class ASCIITreeMaker(ExpressionDescender):
    """
    An expression-descender that is special-purposed for displaying
    s-expressions in an ASCII tree format.
    """
    indent = ' ' * 4

    def __init__(self):
        self.initial = True
        self.initialPrefix = '   '
        self.prefix = '+--'

    def handleExpression(self, sexpr, padding=''):
        return self.descend(sexpr, padding=padding+self.indent)

    def handleSymbol(self, node, padding=''):
        return padding + self.indent + '|--[' + str(node) + ']\n'

    def handleOperator(self, node, padding=''):
        if self.initial:
            self.initial = False
            prefix = self.initialPrefix
        else:
            prefix = self.prefix
        return padding + prefix + '[' + str(node) + ']\n'


class TreeMaker(ExpressionDescender):
    """
    This class descends through an expression and constructs a graph-theoretic
    model of it.

    ASCIITreeMaker graph works easily because parent and child relationships
    aren't actually tracked... only "depth" and it's the depth that's passed
    from descend to handler to descend.

    Here, we are actually quite interested in relationships. We need them, in
    fact, to set up edges.

    In particular, when a list object is found, we want that list object's
    operator (the first element in the list) to know that it's parent is the
    same parent as the list operator's parent, as the symbols that are the list
    operator's siblings, and that all the non-list, non-operators in the in the
    list have the list's operator object as *their* parent.

    For example, if we have the expression (+ 2 3 (* 4 5) 6), then the
    following is true:
       1. + is the root
       2. 2, 3, *, and 6 are its children
       3. 3 and 5 are the children of *

    When the subexpression (* 4 5) is reached in the descent, it needs to know
    that its parent is +.

    What's more, when it has finished with (* 4 5) and has come back up to 6, 6
    needs to know that its parent is not * like 4 and 5, but that it's +.
    """
    def __init__(self):
        self.initial = True
        self.indexCount = 0
        self.graph = graph.Tree()
        self.graph.terminalCount = 0
        self.graph.operatorCount = 0

    def handleExpression(self, sexpr, parentID=0):
        # handle the operator, the first element of the expression
        operator = sexpr[0]
        vertex = graph.Vertex(self.indexCount, operator)
        self.graph.add_node(vertex)
        #print parentID, self.indexCount, operator
        self.graph.add_edge(parentID, vertex.id)
        self.graph.operatorCount += 1
        self.indexCount += 1
        # then handle the expression itself, passing the parentID
        self.descend(sexpr, parentID=vertex.id)
        return []

    def handleSymbol(self, node, parentID=0):
        #print parentID, self.indexCount, node
        vertex = graph.Vertex(self.indexCount, node)
        self.graph.add_node(vertex)
        self.graph.add_edge(parentID, vertex.id)
        self.indexCount += 1
        self.graph.terminalCount += 1

    def handleOperator(self, node, parentID=0):
        """
        We're going to handle operators in the expression handler, since that's
        the only way for operator children to know about their parent. We only
        need to use this method once: when it's called for the first time in
        the descend method.
        """
        if self.initial:
            self.initial = False
            vertex = graph.Vertex(self.indexCount, node)
            self.graph.add_node(vertex)
            #print parentID, self.indexCount, node
            self.graph.operatorCount += 1
            self.indexCount += 1


class ExpressionMaker(ExpressionDescender):
    """
    Given a tree that represents an expression, create an expression from all
    the nodes.

    >>> orig = '(+ 1 2)'
    >>> s = SExpression(orig)
    >>> g = s.getGraph()
    >>> e = ExpressionMaker().descend(g)
    >>> ''.join(e) == orig
    True

    >>> orig = '(+ 1 2 (* 5 3))'
    >>> s = SExpression(orig)
    >>> g = s.getGraph()
    >>> e = ExpressionMaker().descend(g)
    >>> ''.join(e) == orig
    True

    >>> orig = '(+ 1 2 (* 5 3) 4)'
    >>> s = SExpression(orig)
    >>> g = s.getGraph()
    >>> e = ExpressionMaker().descend(g)
    >>> ''.join(e) == orig
    True

    >>> orig = '(+ 1 2 (* 5 3 (- 4 7)))'
    >>> s = SExpression(orig)
    >>> g = s.getGraph()
    >>> e = ExpressionMaker().descend(g)
    >>> ''.join(e) == orig
    True
    """
    def __init__(self):
        self.initial = True

    def handleSymbol(self, symbol):
        return " %s" % str(symbol)

    def handleOperator(self, operator):
        return str(operator)

    def descend(self, tree, parent=0):
        output = []
        if not tree.nodes():
            return output
        if self.initial:
            self.initial = False
        else:
            output.append(" ")
        output.append("(")
        output.append(self.handleOperator(tree[parent].object))
        for childID in sorted(tree.successors(parent)):
            vertex = tree[childID]
            node = vertex.object
            symbolType = lisp.getType(node)
            grandChildren = tree.successors(childID)
            #print node, vertex, symbolType, grandChildren
            if grandChildren:
                output.extend(self.descend(tree, childID))
            elif symbolType == symbol.SymbolObject:
                output.append(self.handleSymbol(node))
        output.append(")")
        return output


def getExpression(obj):
    """
    Get a string representation of an s-expression given a graph-theoretic
    representation.

    >>> orig = '(+ 1 2 (* 5 3 (- 4 7)))'
    >>> s = SExpression(orig)
    >>> e = getExpression(s.getGraph())
    >>> type(e)
    <type 'str'>
    >>> e == orig
    True
    """
    if isinstance(obj, graph.Tree):
        parts = ExpressionMaker().descend(obj)
        return "".join(parts)
    else:
        raise TypeError, "Unexpected instance '%s'." % obj


def getSExpression(obj):
    """
    Given a graph-theoretic representation of an s-expression, return an
    SExpression instance based on it.

    >>> orig = '(+ 1 2 (* 5 3 (- 4 7)))'
    >>> s1 = SExpression(orig)
    >>> s2 = getSExpression(s1.getGraph())
    >>> type(s2)
    <class 'pylispng.sexpr.SExpression'>
    >>> repr(s1) == repr(s2) == orig
    True
    """
    expr = getExpression(obj)
    if expr:
        return SExpression(expr)


class SExpression(object):
    """
    A wrapper class for the ListObject, Parser, and Interpreter. This class
    allows one to build Lisp expressions, one node at a time. Particularly
    useful for genetic programming.


    >>> exprs = [
    ...   '(+ 1 2 (* 5 3))',
    ...   '(+ 1 2 (* 5 3) 4)',
    ...   '(+ 1 2 (* 5 3 ( * (- 4 7) (* 14 3))) 4)',
    ...   ]
    >>> for e in exprs:
    ...   s = SExpression(e)
    ...   print s, "==", list(s), "evals to", s.eval()
    (+ 1 2 (* 5 3)) == [+, <number 1>, <number 2>, *, <number 5>, <number 3>] evals to 18
    (+ 1 2 (* 5 3) 4) == [+, <number 1>, <number 2>, *, <number 5>, <number 3>, <number 4>] evals to 22
    (+ 1 2 (* 5 3 (* (- 4 7) (* 14 3))) 4) == [+, <number 1>, <number 2>, *, <number 5>, <number 3>, *, -, <number 4>, <number 7>, *, <number 14>, <number 3>, <number 4>] evals to -1883

    >>> s1 = SExpression()
    >>> s1.append('+')
    >>> s1.append('2')
    >>> s1.append('4')

    >>> s2 = SExpression('(+ 2 4)')
    >>> s1.eval() == s2.eval()
    True

    >>> list(s1.sexpr)
    [+, <number 2>, <number 4>]
    >>> lisp.getType(s1.sexpr[0])
    <class 'pylispng.symbol.FunctionObject'>

    Let's check to make sure we get the expected number of vertices:

    >>> s1.getGraph().order()
    3

    And then a check for the number of the graph's edges:

    >>> s1.getGraph().size()
    2

    Now let's put another operator in the expression:

    >>> s3 = SExpression()
    >>> s3.append('-')
    >>> s3.append('3')
    >>> s3.append('1')
    >>> s1.append(s3)
    >>> s1.sexpr
    (+ 2 4 (- 3 1))

    >>> s1.getGraph().order()
    6
    >>> s1.getGraph()[0]
    <Vertex 0 (+)>
    >>> s1.getGraph()[3]
    <Vertex 3 (-)>
    """
    def __init__(self, expr=""):
        """

        """
        self.depth = 0
        self.graph = None
        self.depths = util.DynamicList()
        self.sexpr = None
        self.parser = None
        self.setUpParser(expr)

    def __len__(self):
        return len(self.sexpr)

    def __iter__(self):
        return iter(self.parser)

    def __getitem__(self, index):
        """
        >>> s = SExpression('(- 4 8)')
        >>> s[0]
        -
        >>> s[1].value
        4
        >>> s[2].value
        8

        Negative indices are not defined:

        >>> s[-1]
        Traceback (most recent call last):
            ...
        GraphIndexError
        """
        if not self.graph:
            self.makeGraph()
        node = self.getGraph()[index]
        if not node:
            raise GraphIndexError
        return node.object

    def __repr__(self):
        return repr(self.sexpr)

    def setUpParser(self, expr):
        self.parser = parser.Parser(str(expr))
        if expr:
            self.sexpr = self.parser.sexpr
            self.makeGraph()
        else:
            self.sexpr = symbol.ListObject()

    def getSize(self):
        """
        The size of the expression is determined by the number of operators
        (functions) it has.
        """
        return self.getGraph().operatorCount

    def append(self, expr):
        """
        Define append such that when it is called, another s-expression is
        added under the top-most operator in the original s-expression.

        >>> s = SExpression()
        >>> len(s)
        0
        >>> s.append('+')
        >>> len(s)
        1
        >>> s.append('1')
        >>> len(s)
        2
        >>> s.append('2')
        >>> len(s)
        3
        """
        p = parser.Parser(str(expr))
        self.sexpr.append([p.sexpr])
        self.makeGraph()

    def eval(self):
        """
        The lisp environment is needed to do evaluations; this is a convenience
        method which sets that up and then evaluate the s-expression.

        >>> s = SExpression('(+ 1 2)')
        >>> s.eval()
        <number 3>
        >>> s = SExpression()
        >>> s.append('+')
        >>> s.append('1')
        >>> s.append('2')
        >>> s.eval()
        <number 3>
        """
        return lisp.Interpreter().eval(self.sexpr)

    def setDepths(self, sexpr=None, level=0):
        """
        The depths of an expression get set by iterating through each node and
        determining the depth at that point.
        """
        depth = 0
        if not sexpr:
            sexpr = self.sexpr
        for node in sexpr:
            if isinstance(node, symbol.ListObject):
                depth = self.setDepths(node, level + 1)
            else:
                depth = level + 1
            self.depths[depth] = 1
        return depth

    def getDepth(self):
        """
        The depth of an expression is a measure based on the number of nodes
        that must be traversed in order to pass from the top-most, root node to
        the deepest node in the expression.
        """
        # XXX this implementation doesn't match docstring; should traverse the
        # graph instead
        self.depths = util.DynamicList()
        self.setDepths()
        return len(self.depths) - 1

    def hasChildren(self, nodeID):
        """
        >>> s = SExpression('(+ 1 2 (* 5 3) 4)')
        >>> s.hasChildren(0)
        True
        >>> s.hasChildren(1)
        False
        >>> s.hasChildren(2)
        False
        >>> s.hasChildren(3)
        True
        >>> s.hasChildren(4)
        False
        >>> s.hasChildren(5)
        False
        >>> s.hasChildren(6)
        False
        """
        if self.getGraph().successors(nodeID):
            return True
        return False

    def getSubGraph(self, nodeID=0, reindex=True):
        """
        Given a node (vertex) index for a graph representation of the
        expression, return the subgraph that has as its root the node
        associated with the given node ID.

        >>> e = '(+ 1 2 (* 5 3))'
        >>> s = SExpression(e)
        >>> g1 = s.getSubGraph(0, reindex=False)
        >>> g2 = s.getGraph()
        >>> sorted(g1.nodes()) == sorted(g2.nodes())
        True

        >>> g = s.getSubGraph(3)
        >>> sorted(g.nodes())
        [0, 1, 2]
        >>> sorted(g.edges())
        [(0, 1), (0, 2)]

        >>> e = '(+ 1 2 (* 5 3) 4)'
        >>> s = SExpression(e)
        >>> g1 = s.getSubGraph(0, reindex=False)
        >>> g2 = s.getGraph()
        >>> sorted(g1.nodes()) == sorted(g2.nodes())
        True

        >>> g = s.getSubGraph(3)
        >>> sorted(g.nodes())
        [0, 1, 2]
        >>> sorted(g.edges())
        [(0, 1), (0, 2)]

        >>> e = '(* 5 (+ 3 4 6) 22 (* 2.4 5 0.001) (- 5 (+ 2 (- 2 4))))'
        >>> s = SExpression(e)

        >>> g = s.getSubGraph(2)
        >>> g.nodes()
        [0, 1, 2, 3]
        >>> g.edges()
        [(0, 1), (0, 2), (0, 3)]

        >>> g = s.getSubGraph(7)
        >>> g.nodes()
        [0, 1, 2, 3]
        >>> g.edges()
        [(0, 1), (0, 2), (0, 3)]

        >>> g = s.getSubGraph(11)
        >>> g.nodes()
        [0, 1, 2, 3, 4, 5, 6]
        >>> g.edges()
        [(0, 1), (0, 2), (2, 3), (2, 4), (4, 5), (4, 6)]
        """
        nodes = self.getGraph().get_sub_nodes(nodeID)
        return self.getGraph().get_subgraph(nodes, reindex=reindex)

    def getSubExpression(self, nodeID=0, reindex=True):
        """
        >>> e = '(+ 1 2 (* 5 3))'
        >>> s = SExpression(e)
        >>> s.getSubExpression(0, reindex=False)
        (+ 1 2 (* 5 3))
        >>> s.getSubExpression(3)
        (* 5 3)

        >>> e = '(+ 1 2 (* 5 3) 4)'
        >>> s = SExpression(e)
        >>> s.getSubExpression(0, reindex=False)
        (+ 1 2 (* 5 3) 4)
        >>> s.getSubExpression(3)
        (* 5 3)

        >>> e = '(+ 1 2 (* 5 3 ( * (- 4 7) (* 14 3))) 4)'
        >>> s = SExpression(e)
        >>> s.getSubExpression(3)
        (* 5 3 (* (- 4 7) (* 14 3)))

        >>> s.getSubExpression(6)
        (* (- 4 7) (* 14 3))

        >>> s.getSubExpression(10)
        (* 14 3)

        >>> e = '(* 5 (+ 3 4 6) 22 (* 2.4 5 0.001) (- 5 (+ 2 (- 2 4))))'
        >>> s = SExpression(e)

        >>> s[0]
        *
        >>> s.getSubExpression(0)
        (* 5 (+ 3 4 6) 22 (* 2.4 5 0.001) (- 5 (+ 2 (- 2 4))))

        >>> s[2]
        +
        >>> s.getSubExpression(2)
        (+ 3 4 6)

        >>> s[7]
        *
        >>> s.getSubExpression(7)
        (* 2.4 5 0.001)

        >>> s[11]
        -
        >>> s.getSubExpression(11)
        (- 5 (+ 2 (- 2 4)))
        """
        g = self.getSubGraph(nodeID, reindex=reindex)
        return getSExpression(g)

    def setSubExpression(self, nodeID, sexpr):
        """
        >>> e = '(+ 1 2 (* 5 3))'
        >>> s = SExpression(e)
        >>> s.getSubExpression(3)
        (* 5 3)

        >>> sexpr = '(- 4 7)'
        >>> s.setSubExpression(3, sexpr)
        >>> s
        (+ 1 2 (- 4 7))

        >>> sexpr = SExpression('(- 4 7)')
        >>> s.setSubExpression(3, sexpr)
        >>> s
        (+ 1 2 (- 4 7))

        >>> sexpr = SExpression('(- 4 7)')
        >>> s.setSubExpression(3, sexpr.getGraph())
        >>> s
        (+ 1 2 (- 4 7))
        """
        if isinstance(sexpr, graph.Tree):
            subgraph = sexpr
        elif isinstance(sexpr, SExpression):
            subgraph = sexpr.getGraph()
        elif True in [isinstance(sexpr, x) for x in [str, unicode]]:
            sexpr = SExpression(sexpr)
            subgraph = sexpr.getGraph()
        g = self.getGraph()
        g.replace_subgraph(nodeID, subgraph)
        #import pdb
        #pdb.set_trace()
        self.setUpParser(getExpression(g))
        #pdb.set_trace()

    def makeGraph(self):
        """
        Descend the expression, building a graph with operators, functions,
        numbers, etc., as the vertices on the graph.
        """
        gm = TreeMaker()
        gm.descend(self.sexpr)
        self.graph = gm.graph

    def getGraph(self):
        if not self.graph:
            self.makeGraph()
        return self.graph

    def printTree(self):
        """
        """
        sprint(self)


def getTree(sexpr):
    """
    An s-expression tree builder.
    """
    if isinstance(sexpr, SExpression):
        sexpr = sexpr.sexpr
    elif isinstance(sexpr, graph.Tree):
        sexpr = getSExpression(sexpr).sexpr
    tree = ASCIITreeMaker()
    results = tree.descend(sexpr, padding='')
    return ''.join(results)


def sprint(sexpr):
    print '\n', getTree(sexpr)


if __name__ == '__main__':
    import doctest
    doctest.testmod()
