"""
A trivial Lisp implementation.
"""
import re
import sys
import types
import string

from pylispng.util import DynamicList, property_cache
from pylispng.exception import Error, UnboundSymbolError


class Evalable(object):
    """
    Evaluatable Lisp object interface.

    The type system: every type knows how to evaluate itself but must be handed
    an Environment to work with, plus any arguments it might take.
    """
    def eval(self, environment, args=None):
        raise Error, "Unimplemented"


class SymbolObject(Evalable):
    """
    Everything in Lisp is a list of symbols.
    """
    def __init__(self, name):
        self.name = name
        if not property_cache.has_key(self.name):
            property_cache[self.name] = {}

    def __repr__(self):
        return self.name

    def __hash__(self):
        return hash(self.name)

    def eval(self, env, args=None):
        return env.get(self.name)

    def cons(self, item):
        return PairObject(item, self)

    def nullp(self):
        return FALSE

    # property lists (really a hash table here).
    def put(self, property, value):
        property_cache[self.name][property] = value

    def get(self, property):
        try:
            return property_cache[self.name][property]
        except:
            return FALSE


class StringObject(Evalable):

    def __init__(self, str):
        if isinstance(str, StringObject):
            str = str.string
        self.string = str

    def __repr__(self):
        return `self.string`

    def eval(self, env, args=None):
        return self.string

    def nullp(self):
        if self.string == "":
            return TRUE
        else:
            return FALSE

    def __getitem__(self, index):
        return self.string[index]


class NumberObject(Evalable):
    """
    Test string-to-numbers:
    >>> NumberObject('1')
    <number 1>
    >>> NumberObject('1').value
    1
    >>> NumberObject('1.7').value
    1.7
    >>> NumberObject('1' * 20).value
    11111111111111111111L

    Test number values:
    >>> NumberObject(1).value
    1
    >>> NumberObject(1.7).value
    1.7
    >>> NumberObject(2**20).value
    1048576

    Test something that can't be coerced:
    >>> NumberObject('1.ch')
    Traceback (most recent call last):
    ...
    ValueError: Cannot be coerced to a number
    """
    def __init__(self, value):
        self.value = None
        self.setValue(value)
        if self.value == None:
            raise ValueError, "Cannot be coerced to a number"

    def __repr__(self):
        return "<number %s>" % self.value

    def __str__(self):
        return str(self.value)

    def __cmp__(self, other):
        # if isinstance(other, NumberObject) or isinstance(other, LogicObject):
        if type(self) == type(other):
            if self.value == other.value:
                return 0
            elif self.value > other.value:
                return 1
            else:
                return -1
        else:
            if self.value == other:
                return 0
            elif self.value > other:
                return 1
            else:
                return -1

    def __add__(self, other):
        if type(self) == type(other):
            return NumberObject(self.value + other.value)
        else:
            return self.value + other
    __radd__ = __add__

    def __sub__(self, other):
        if type(self) == type(other):
            return NumberObject(self.value - other.value)
        else:
            return self.value - other
    __rsub__ = __sub__

    def __mul__(self, other):
        if type(self) == type(other):
            return NumberObject(self.value * other.value)
        else:
            return self.value * other
    __rmul__ = __mul__

    def __div__(self, other):
        if type(self) == type(other):
            return NumberObject(self.value / float(other.value))
        else:
            return self.value / float(other)
    __rdiv__ = __div__

    def __mod__(self, other):
        if type(self) == type(other):
            return NumberObject(self.value % other.value)
        else:
            return self.value % other
    __rmod__ = __mod__

    def eval(self, env, args=None):
        return self

    def nullp(self):
        if self.value == 0.0:
            return TRUE
        else:
            return FALSE

    def setValue(self, value):
        INT = re.compile(r'^[+-]?\d+$')
        LONGINT = re.compile(r'^[+-]?\d+[lL]$')
        FLOAT = re.compile(r'^[+-]?(\d+\.\d*$|\d*\.\d+$)')
        stringToNumMap = { 
            INT: string.atoi,
            LONGINT: string.atol,
            FLOAT: string.atof
            }
        if isinstance(value, NumberObject):
            self.value = value.value
        elif True in [isinstance(value, x) for x in [int, float, long]]:
            self.value = value
        elif True in [isinstance(value, x) for x in [str, unicode]]:
            for number in INT, LONGINT, FLOAT:
                if number.match(value):
                    self.value = stringToNumMap[number](value)

    def getValue(self):
        return self.value


class LogicObject(Evalable):

    def __init__(self, value):
        if isinstance(value, NumberObject):
            self.value = value.value
        else:
            self.value = value

    def __repr__(self):
        if self.value == 1.0:
            return '*true*'
        elif self.value == 0.0:
            return '*false*'
        else:
            return "(logic %s)" % self.value

    def __cmp__(self, other):
        #if type(self) == type(other):
        if isinstance(other, LogicObject):
            if self.value == other.value:
                return 0
            elif self.value > other.value:
                return 1
            else:
                return -1
        else:
            if self.value == other:
                return 0
            elif self.value > other:
                return 1
            else:
                return -1

    def __neg__(self):
        return LogicObject(1.0 - self.value)

    def __and__(self, other):
        return LogicObject(min(self.value, other.value))

    def __or__(self, other):
        return LogicObject(max(self.value, other.value))

    def eval(self, env, args=None):
        return self

    def nullp(self):
        return -self


TRUE = LogicObject(1.0)
FALSE = LogicObject(0.0)


class PairObject(Evalable):
    """
    Not quite a list. In real Lisps, all lists are just piled up pairs.
    Not so in PyLisp. Pairs are special case, mostly for association
    lists and the sort.
    """
    def __init__(self, a, d):
        self.a, self.d = a, d

    def __repr__(self):
        if self.d == None:
            return "(%s)" % (self.a, self.d)
        else:
            return "(%s . %s)" % (self.a, self.d)

    def replaca(self, val):
        self.a = val
        return val

    def replacd(self, val):
        if isinstance(val, ListObject):
            return val.cons(self.a)
        else:
            self.d = val
            return val

    def first(self):
        return self.a

    def rest(self):
        return self.d

    def second(self):
        return self.d

    def nullp(self):
        return FALSE


class ListObject(Evalable):

    def __init__(self, lst=None):
        if lst is None:
            self.list = []
        else:
            self.list = lst

    def __repr__(self):
        if self.list == []:
            return "()"
        s = "(%s" % self.list[0]
        for e in self.list[1:]:
            s = s + " %s" % e
        return s + ")"

    def __len__(self):
        return len(self.list)

    def __getitem__(self, i):
        return self.list[i]

    def __and__(self, other):
        if isinstance(other, LogicObject):
            if self.nullp():
                return FALSE & other
            else:
                return TRUE & other

    def first(self):
        return self.list[0]

    def replaca(self, val):
        self.list[0] = val
        return val

    def second(self):
        try:
            return self.list[1]
        except:
            return ListObject([])

    def third(self):
        try:
            return self.list[2]
        except:
            return ListObject([])

    def rest(self):
        try:
            return ListObject(self.list[1:])
        except:
            return ListObject([])

    def replacd(self, val):
        self.list = [self.list[0]] + val.list
        return val

    def cons(self, item):
        n = ListObject(self.list[:])
        n.list.insert(0, item)
        return n

    def append(self, lst):
        if isinstance(lst, ListObject):
            #return ListObject(self.list + lst.list)
            self.list = self.list + lst.list
        else:
            #return ListObject(self.list + lst)
            self.list = self.list + lst

    def nconc(self, lst):
        self.append(lst)
        return self

    def eval(self, env, args=None):
        """
        Code and data look the same in lisp.
        """
        fn = self.list[0].eval(env)
        if isinstance(fn, LambdaObject) or \
           isinstance(fn, MacroObject) or \
           isinstance(fn, FunctionObject) or \
           isinstance(fn, SyntaxObject):
            return fn.eval(env, self.list[1:])
        else:
            raise Error, "%s is not applyable" % fn

    def nullp(self):
        if len(self.list) == 0:
            return TRUE
        else:
            return FALSE

    def iquote(self, env):
        """
        l.iquote(environment) - do the back-tick substitutions
        """
        answer = []
        length = len(self.list)
        i = 0
        while i < length:
            if isinstance(self.list[i], ListObject):
                answer.append(self.list[i].iquote(env))
            elif type(self.list[i]) == types.StringType:
                if self.list[i] == ',':
                    i = i + 1
                    if type(self.list[i]) == types.StringType:
                        if self.list[i] == '@':
                            # splice: `(a b ,@fred) -> (a b a b c)
                            i = i + 1
                            answer = answer + self.list[i].eval(env).list
                        else:
                            answer.append(self.list[i].eval(env))
                    else:
                        answer.append(self.list[i].eval(env))
            else:
                answer.append(self.list[i])
            i = i + 1
        return ListObject(answer)

    def assoc(self, item):
        for pair in self.list:
            if isinstance(pair, PairObject) or isinstance(pair, ListObject):
                key = pair.first()
                if (isinstance(key, SymbolObject) and 
                    isinstance(item, SymbolObject)):
                    if key.name == item.name:
                        return pair.second()
                else:
                    if key == item:
                        return pair.second()
            else:
                raise Error(
                    "assoc: second argument must be list of lists/pairs")
        return FALSE


NIL = ListObject([])


class ObjectObject(Evalable):
    """
    A basic message-passing object orientation for PyLisp
    """
    def __init__(self, name, *attrs, **opts):
        self.name = name

    def __repr__(self):
        return "<object %s>" % id(self.name)

    def getName(self):
        return self.name


class FunctionObject(Evalable):
    """
    Internal (Python) functions
    """
    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return "<function '%s'>" % self.name

    def eval(self, env, args):
        return self.name(env, args)

    def nullp(self):
        return FALSE

    def getName(self):
        return self.name


class LambdaObject(Evalable):
    """
    Fuctions defined in PyLisp.
    """
    def __init__(self, bindings, code, env):
        self.bindings = bindings
        self.code = code
        # Remember the defining environment.
        self.env = env

    def __repr__(self):
        if len(self.code) == 1:
            return "(lambda %s %s)" % (self.bindings, self.code[0])
        else:
            return "(lambda %s %s)" % (
                self.bindings, ListObject(['begin'] + self.code))

    def eval(self, env, args):
        interpreter = env.get('*self*')
        # If the lexical environment in which this was defined is different
        # than the one called from, push that saved environment's values
        # onto the new lambda's environment.
        if self.env:
            interpreter.pushE(self.env.e)
        else:
            interpreter.pushE()
        # Evaluate the arguments and bind to the correct names.
        i = 0
        n = len(self.bindings)
        while i < n:
            if self.bindings[i].name == "&rest":
                rest = []
                for arg in args[i:]:
                    rest.append(arg.eval(interpreter.e))
                interpreter.lexicalIntern(
                    self.bindings[i + 1].name, ListObject(rest))
                i = i + 1
            else:
                interpreter.lexicalIntern(
                    self.bindings[i].name, args[i].eval(interpreter.e))
            i = i + 1
        # Run the code.
        ans = FALSE
        for code in self.code:
            ans = code.eval(interpreter.e)
        interpreter.popE()
        return ans

    # Mostly transform things like the let into the lambda form, which
    # will speed things up a bit.
    def compile(self):
        pass

    def nullp(self):
        return FALSE


class MacroObject(Evalable):
    # (def incf (macro (x) `(+ ,x 1)))
    # ((macro (x) `(+ ,x 1)) swee)
    def __init__(self, bindings, code, env):
        self.bindings = bindings
        self.code = code
        self.env = env                  # defining environment

    def __repr__(self):
        if len(self.code) == 1:
            return "(macro %s %s)" % (self.bindings, self.code[0])
        else:
            # ???
            return "(macro %s %s)" % (
                self.bindings, ListObject(['begin'] + self.code))

    def eval(self, env, args):
        interpreter = env.get('*self*')
        # If the lexical environment in which this was defined is different
        # than the one called from, push that saved environment's values
        # onto the new macro's environment.
        if self.env:
            interpreter.pushE(self.env.e)
        else:
            interpreter.pushE()
        # Evaluate the arguments and bind to the correct names: UNEVALUATED.
        i = 0
        n = len(self.bindings)
        while i < n:
            if self.bindings[i].name == "&rest":
                interpreter.lexicalIntern(
                    self.bindings[i + 1].name, ListObject(args[i:]))
                i = i + 1
            else:
                interpreter.lexicalIntern(self.bindings[i].name, args[i])
            i = i + 1
        # Run the code.  The first run to expand the macro.
        mcode = []
        for mc in self.code:
            mcode.append(mc.eval(interpreter.e))
        # The macro bindings environment is finished.
        interpreter.popE()
        # Now, run again, this time for real.
        ans = FALSE
        for code in mcode:
            ans = code.eval(interpreter.e)
        return ans

    def expand(self, env, args):
        interpreter = env.get('*self*')
        if self.env:
            interpreter.pushE(self.env.e)
        else:
            interpreter.pushE()
        # Evaluate the arguments and bind to the correct names: UNEVALUATED.
        i = 0
        n = len(self.bindings)
        while i < n:
            if self.bindings[i].name == "&rest":
                interpreter.lexicalIntern(
                    self.bindings[i + 1].name, ListObject(args[i:]))
                i = i + 1
            else:
                interpreter.lexicalIntern(self.bindings[i].name, args[i])
            i = i + 1
        mcode = []
        for mc in self.code:
            mcode.append(mc.eval(interpreter.e))
        interpreter.popE()
        # The 'begin' turns the (Python) list of code into legally
        # parsable list of Pylisp code.
        return ListObject(mcode).cons('begin')

    def nullp(self):
        return FALSE


class SyntaxObject(Evalable):
    """
    Same as the FunctionObject, but this object is expected to handle
    the evaluation of the arguments it is passed.
    """
    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return "<function '%s'>" % self.name

    def eval(self, env, args):
        return self.name(env, args)

    def nullp(self):
        return FALSE


class SystemObject(Evalable):
    pass


# some syntactic sugar for the ' symbol.
QUOTE = SymbolObject('quote')
# and for `
IQUOTE = SymbolObject('i-quote')


mapper = {
    '*self*': SystemObject,
    '*env*': SystemObject,
    '*true*': TRUE,
    '*false*': FALSE,
    '*true-enough*': TRUE,
    '*gensym-string*': SymbolObject, #'GENSYM@', 'no_self',
    '*gensym-counter*': NumberObject, #0, 'no_self',
    'print': FunctionObject,
    'first': FunctionObject,
    'car': FunctionObject,
    'replaca': FunctionObject,
    'second': FunctionObject,
    'cadr': FunctionObject,
    'third': FunctionObject,
    'rest': FunctionObject,
    'cdr': FunctionObject,
    'replacd': FunctionObject,
    'cons': FunctionObject,
    'append': FunctionObject,
    'null?': FunctionObject,
    'list': FunctionObject,
    'assoc': FunctionObject,
    '+': FunctionObject,
    '-': FunctionObject,
    '/': FunctionObject,
    '%': FunctionObject,
    '*': FunctionObject,
    '>': FunctionObject,
    '>=': FunctionObject,
    '<': FunctionObject,
    '<=': FunctionObject,
    '==': FunctionObject,
    'eql': FunctionObject,
    '!=': FunctionObject,
    'not': FunctionObject,
    'and': FunctionObject,
    'or': FunctionObject,
    'logic': FunctionObject,
    'load': FunctionObject,
    'gensym': FunctionObject,
    'eval': FunctionObject,
    'list?': FunctionObject,
    'pair?': FunctionObject,
    'symbol?': FunctionObject,
    'string?': FunctionObject,
    'number?': FunctionObject,
    'logic?': FunctionObject,
    'symbol-name': FunctionObject,
    'top-level': FunctionObject,
    'the-environment': FunctionObject,
    'get-environment': FunctionObject,
    'env-get': FunctionObject,
    'env-set': FunctionObject,
    'get': FunctionObject,
    'put': FunctionObject,
    'elt': FunctionObject,
    'apply': FunctionObject,
    'send': FunctionObject,
    # Chat with Python.
    #'py-setattr': [symbol.FunctionObject, 'do_py_setattr'],
    #'py-getattr': [symbol.FunctionObject, 'do_py_getattr'],
    #'py-hash-make': [symbol.FunctionObject, 'do_py_hash_make'],
    #'py-hash-get': [symbol.FunctionObject, 'do_py_hash_get'],
    #'py-hash-set': [symbol.FunctionObject, 'do_py_hash_set'],
    #'py-send': [symbol.FunctionObject, 'do_py_send'],
    'py-format': FunctionObject,
    'py-type': FunctionObject,
    'py-eval': FunctionObject,
    'py-exec': FunctionObject,
    'keyword->pyhash': FunctionObject,
    # macros/syntax
    'begin': SyntaxObject,
    'setq': SyntaxObject,
    'def': SyntaxObject,
    'let': SyntaxObject,
    'lambda': SyntaxObject,
    'macro': SyntaxObject,
    'macro-expand': SyntaxObject,
    'quote': SyntaxObject,
    'i-quote': SyntaxObject,
    'if': SyntaxObject,
    'cond': SyntaxObject,
    'with-py-hash-env': SyntaxObject,
    }

if __name__ == "__main__":
    import doctest
    doctest.testmod()
