#!/usr/bin/env python
import copy
import unittest

from pylispng.graph import Graph, DiGraph
from pylispng.exception import GraphError


class GraphTestCase(unittest.TestCase):

    def setUp(self):
        self.Graph=Graph
        # build K3
        self.k3adj={0: {1: 1, 2: 1}, 1: {0: 1, 2: 1}, 2: {0: 1, 1: 1}}
        self.k3edges=[(0, 1), (0, 2), (1, 2)]
        self.k3nodes=[0, 1, 2]
        self.K3=self.Graph()
        self.K3.adj=self.k3adj


    def test_name(self):
        G=self.Graph(name='')
        self.assertEquals(G.name,"")
        G=self.Graph(name='test')
        self.assertEquals(G.__str__(),"test")
        self.assertEquals(G.name,"test")

    def test_data_input(self):
        G=self.Graph(name="test")
        G.add_nodes_from([1, 2])
        G.add_edge(1,2)
        G.add_edge(2,1)
        self.assertEquals(G.name,"test")
        self.assertEquals(sorted(G.adj.items()),[(1, {2: 1}), (2, {1: 1})])


    def test_contains(self):
        G=self.K3
        self.assertTrue(1 in G )
        self.assertTrue(4 not in G )
        self.assertTrue('b' not in G )
        self.assertTrue([] not in G )   # no exception for nonhashable
        self.assertTrue({1:1} not in G) # no exception for nonhashable

    def test_order(self):
        G=self.K3
        self.assertEquals(len(G),3)
        self.assertEquals(G.order(),3)
        self.assertEquals(G.number_of_nodes(),3)

    def test_getitem(self):
        G=self.K3
        self.assertEquals(G[0],{1: 1, 2: 1})
        self.assertRaises(KeyError, G.__getitem__, 'j')
        self.assertRaises((TypeError,GraphError), G.__getitem__, ['A'])

    def test_add_node(self):
        G=self.Graph()
        G.add_node(0)
        self.assertEquals(G.adj,{0:{}})
    
    def test_add_nodes_from(self):
        G=self.Graph()
        G.add_nodes_from([0,1,2])
        self.assertEquals(G.adj,{0:{},1:{},2:{}})

    def test_remove_node(self):
        G=self.K3
        G.remove_node(0)
        self.assertEquals(G.adj,{1:{2:1},2:{1:1}})
        self.assertRaises((KeyError,GraphError), G.remove_node,-1)

        # generator here to implement list,set,string...
    def test_remove_nodes_from(self):
        G=self.K3
        G.remove_nodes_from([0,1])
        self.assertEquals(G.adj,{2:{}})
        G.remove_nodes_from([-1]) # silent fail


    def test_nodes_iter(self):
        G=self.K3
        self.assertEquals(sorted(G.nodes_iter()),self.k3nodes)

    def test_nodes(self):
        G=self.K3
        self.assertEquals(sorted(G.nodes()),self.k3nodes)

    def test_has_node(self):
        G=self.K3
        self.assertTrue(G.has_node(1))
        self.assertTrue(not G.has_node(4))
        self.assertTrue(not G.has_node([]))   # no exception for nonhashable
        self.assertTrue(not G.has_node({1:1})) # no exception for nonhashable


    def test_add_edge(self):
        G=self.Graph()
        G.add_edge(0,1)
        self.assertEquals(G.adj,{0: {1: 1}, 1: {0: 1}})
        G=self.Graph()
        G.add_edge(*(0,1))
        self.assertEquals(G.adj,{0: {1: 1}, 1: {0: 1}})
    
    def test_add_edges_from(self):
        G=self.Graph()
        G.add_edges_from([(0,1),(0,2)])
        self.assertEquals(G.adj,{0: {1:1, 2:1}, 1: {0:1}, 2:{0:1}})
        G=self.Graph()
        G.add_edges_from([(0,1,3),(0,2)])
        self.assertEquals(G.adj,{0: {1:3, 2:1}, 1: {0:3}, 2:{0:1}})
        G=self.Graph()
        G.add_edges_from([(0,1),(0,2,3)],data=2)
        self.assertEquals(G.adj,{0: {1:2, 2:3}, 1: {0:2}, 2:{0:3}})
        G.add_edges_from([(0,1),(0,2,3)],data=2)
        self.assertRaises(GraphError, G.add_edges_from,[(0,)])  # too few in tuple
        self.assertRaises(GraphError, G.add_edges_from,[(0,1,2,3)])  # too many in tuple
        self.assertRaises(TypeError, G.add_edges_from,[0])  # not a tuple


    def test_remove_edge(self):
        G=self.K3
        G.remove_edge(0,1)
        self.assertEquals(G.adj,{0:{2:1},1:{2:1},2:{0:1,1:1}})
        self.assertRaises((KeyError,GraphError), G.remove_edge,-1,0)

    def test_remove_edges_from(self):
        G=self.K3
        G.remove_edges_from([(0,1)])
        self.assertEquals(G.adj,{0:{2:1},1:{2:1},2:{0:1,1:1}})
        G.remove_edges_from([(0,0)]) # silent fail

    def test_has_edge(self):
        G=self.K3
        self.assertEquals(G.has_edge(0,1),True)
        self.assertEquals(G.has_edge(0,-1),False)

    def test_has_neighbor(self):
        G=self.K3
        self.assertEquals(G.has_neighbor(0,1),True)
        self.assertEquals(G.has_neighbor(0,-1),False)

    def test_neighbors(self):
        G=self.K3
        self.assertEquals(sorted(G.neighbors(0)),[1,2])
        self.assertRaises((KeyError,GraphError), G.neighbors,-1)

    def test_neighbors_iter(self):
        G=self.K3
        self.assertEquals(sorted(G.neighbors_iter(0)),[1,2])
        self.assertRaises((KeyError,GraphError), G.neighbors_iter,-1)

    def test_edges(self):
        G=self.K3
        self.assertEquals(sorted(G.edges()),[(0,1),(0,2),(1,2)])
        self.assertEquals(sorted(G.edges(0)),[(0,1),(0,2)])
        self.assertRaises((KeyError,GraphError), G.neighbors,-1)

    def test_edges_iter(self):
        G=self.K3
        self.assertEquals(sorted(G.edges_iter()),[(0,1),(0,2),(1,2)])
        self.assertEquals(sorted(G.edges_iter(0)),[(0,1),(0,2)])
        self.assertRaises((KeyError,GraphError), G.neighbors_iter,-1)

    def test_get_edge(self):
        G=self.K3
        self.assertEquals(G.get_edge(0,1),1)
        self.assertEquals(G[0][1],1)
        self.assertRaises((KeyError,GraphError), G.get_edge,-1,0)

    def test_adjacency_list(self):
        G=self.K3
        self.assertEquals(G.adjacency_list(),[[1,2],[0,2],[0,1]])

    def test_adjacency_iter(self):
        G=self.K3
        self.assertEquals(dict(G.adjacency_iter()),
                          {0: {1: 1, 2: 1}, 1: {0: 1, 2: 1}, 2: {0: 1, 1: 1}})
                          
    def test_degree(self):
        G=self.K3
        self.assertEquals(G.degree(),[2,2,2])
        self.assertEquals(G.degree(with_labels=True),{0:2,1:2,2:2})
        self.assertEquals(G.degree(0),[2])
        self.assertEquals(G.degree(0,with_labels=True),{0:2})
        self.assertEquals(G.degree(-1),[])

    def test_degree_iter(self):
        G=self.K3
        self.assertEquals(list(G.degree_iter()),[(0,2),(1,2),(2,2)])
        self.assertEquals(dict(G.degree_iter()),{0:2,1:2,2:2})
        self.assertEquals(list(G.degree_iter(0)),[(0,2)])

    def test_clear(self):
        G=self.K3
        G.clear()
        self.assertEquals(G.adj,{})

    def test_copy(self):
        G=self.K3
        H=G.copy()
        self.assertEquals(G.adj,H.adj)

    def test_subgraph(self):
        G=self.K3
        H=G.subgraph([0,1,2])
        self.assertEquals(H.adj,G.adj) 
        H=G.subgraph(0)
        self.assertEquals(H.adj,{0:{}})
        H=G.subgraph([])
        self.assertEquals(H.adj,{})
        H=G.subgraph([],copy=False)
        self.assertEquals(H.adj,{})
        self.assertEquals(G.adj,{})
        self.assertEquals(G.adj,H.adj)

    def test_selfloops(self):
        G=self.K3.copy()
        G.add_edge(0,0)
        self.assertEquals(G.nodes_with_selfloops(),[0])
        self.assertEquals(G.selfloop_edges(data=True),[(0,0,1)])
        self.assertEquals(G.number_of_selfloops(),1)
        G.remove_node(0)
        self.assertEquals(G.order(),2)
        G=self.K3
        G.add_edge(0,0)
        G.add_edge(1,1)
        G.remove_nodes_from([0,1])
        self.assertEquals(G.order(),1)

    def test_size(self):
        G=self.K3
        self.assertEquals(G.size(),3)
        self.assertEquals(G.number_of_edges(),3)

    def test_add_star(self):
        G=self.K3.copy()
        nlist=[12,13,14,15]
        G.add_star(nlist)
        self.assertEquals(sorted(G.edges(nlist)),[(12,13),(12,14),(12,15)])
        G=self.K3.copy()
        data=[2.,3.,4.]
        G.add_star(nlist,data)
        self.assertEquals(sorted(G.edges(nlist,data=True)),[(12,13,2.),(12,14,3.),(12,15,4.)])

    def test_add_path(self):
        G=self.K3.copy()
        nlist=[12,13,14,15]
        G.add_path(nlist)
        self.assertEquals(sorted(G.edges(nlist)),[(12,13),(13,14),(14,15)])
        G=self.K3.copy()
        data=[2.,3.,4.]
        G.add_path(nlist,data)
        self.assertEquals(sorted(G.edges(nlist,data=True)),[(12,13,2.),(13,14,3.),(14,15,4.)])

    def test_add_cycle(self):
        G=self.K3.copy()
        nlist=[12,13,14,15]
        oklists=[ [(12,13),(12,15),(13,14),(14,15)], [(12,13),(13,14),(14,15),(15,12)] ]
        G.add_cycle(nlist)
        self.assertTrue(sorted(G.edges(nlist)) in oklists)
        G=self.K3.copy()
        data=[2.,3.,4.,5.]
        oklists=[ [(12,13,2.),(12,15,5.),(13,14,3.),(14,15,4.)], [(12,13,2.),(13,14,3.),(14,15,4.),(15,12,5.)] ]
        G.add_cycle(nlist,data)
        self.assertTrue(sorted(G.edges(nlist,data=True)) in oklists)

    def test_nbunch_iter(self):
        G=self.K3
        self.assertEquals(list(G.nbunch_iter()),self.k3nodes) # all nodes
        self.assertEquals(list(G.nbunch_iter(0)),[0]) # single node
        self.assertEquals(list(G.nbunch_iter([0,1])),[0,1]) # sequence
        self.assertEquals(list(G.nbunch_iter([-1])),[]) # sequence with none in graph
        self.assertEquals(list(G.nbunch_iter("foo")),[]) # string sequence with none in graph
        bunch=G.nbunch_iter(-1)
        self.assertRaises(StopIteration,bunch.next)
        bunch=G.nbunch_iter([0,1,2,{}])    # unhashable doesn't get caught upon creation of iterator
        self.assertRaises(GraphError,list,bunch) # but gets caught when iterator hits the unhashable


class DiGraphTestCase(GraphTestCase):

    def setUp(self):
        self.Graph=DiGraph
        # build K3
        self.k3adj={0: {1: 1, 2: 1}, 1: {0: 1, 2: 1}, 2: {0: 1, 1: 1}}
        self.k3edges=[(0, 1), (0, 2), (1, 2)]
        self.k3nodes=[0, 1, 2]
        self.K3=self.Graph()
        self.K3.adj=self.k3adj
        self.K3.succ=self.k3adj
        self.K3.pred=copy.deepcopy(self.k3adj)


    def test_data_input(self):
        G=self.Graph(name="test")
        G.add_nodes_from([1, 2])
        G.add_edge(1,2)
        G.add_edge(2,1)
        self.assertEquals(G.name,"test")
        self.assertEquals(sorted(G.adj.items()),[(1, {2: 1}), (2, {1: 1})])
        self.assertEquals(sorted(G.succ.items()),[(1, {2: 1}), (2, {1: 1})])
        self.assertEquals(sorted(G.pred.items()),[(1, {2: 1}), (2, {1: 1})])


    def test_add_edge(self):
        G=self.Graph()
        G.add_edge(0,1)
        self.assertEquals(G.adj,{0: {1: 1}, 1: {}})
        self.assertEquals(G.succ,{0: {1: 1}, 1: {}})
        self.assertEquals(G.pred,{0: {}, 1: {0:1}})
        G=self.Graph()
        G.add_edge(*(0,1))
        self.assertEquals(G.adj,{0: {1: 1}, 1: {}})
        self.assertEquals(G.succ,{0: {1: 1}, 1: {}})
        self.assertEquals(G.pred,{0: {}, 1: {0:1}})

    
    def test_add_edges_from(self):
        G=self.Graph()
        G.add_edges_from([(0,1)])
        self.assertEquals(G.adj,{0: {1: 1}, 1: {}})
        self.assertEquals(G.succ,{0: {1: 1}, 1: {}})
        self.assertEquals(G.pred,{0: {}, 1: {0:1}})


    def test_remove_edge(self):
        G=self.K3
        G.remove_edge(0,1)
        self.assertEquals(G.succ,{0:{2:1},1:{0:1,2:1},2:{0:1,1:1}})        
        self.assertEquals(G.pred,{0:{1:1, 2:1}, 1:{2:1}, 2:{0:1,1: 1}})
        self.assertRaises((KeyError, GraphError), G.remove_edge,-1,0)


    def test_remove_edges_from(self):
        G=self.K3
        G.remove_edges_from([(0,1)])
        self.assertEquals(G.succ,{0:{2:1},1:{0:1,2:1},2:{0:1,1:1}})        
        self.assertEquals(G.pred,{0:{1:1, 2:1}, 1:{2:1}, 2:{0:1,1: 1}})
        G.remove_edges_from([(0,0)]) # silent fail


    def test_copy(self):
        G=self.K3
        H=G.copy()
        self.assertEquals(G.adj,H.adj)
        self.assertEquals(G.succ,H.succ)
        self.assertEquals(G.pred,H.pred)

    def test_has_successor(self):
        G=self.K3
        self.assertEquals(G.has_successor(0,1),True)
        self.assertEquals(G.has_successor(0,-1),False)

    def test_successors(self):
        G=self.K3
        self.assertEquals(sorted(G.successors(0)),[1,2])
        self.assertRaises((KeyError,GraphError), G.successors,-1)

    def test_successors_iter(self):
        G=self.K3
        self.assertEquals(sorted(G.successors_iter(0)),[1,2])
        self.assertRaises((KeyError,GraphError), G.successors_iter,-1)

    def test_has_predecessor(self):
        G=self.K3
        self.assertEquals(G.has_predecessor(0,1),True)
        self.assertEquals(G.has_predecessor(0,-1),False)

    def test_predecessors(self):
        G=self.K3
        self.assertEquals(sorted(G.predecessors(0)),[1,2])
        self.assertRaises((KeyError,GraphError), G.predecessors,-1)

    def test_predecessors_iter(self):
        G=self.K3
        self.assertEquals(sorted(G.predecessors_iter(0)),[1,2])
        self.assertRaises((KeyError,GraphError), G.predecessors_iter,-1)


    def test_edges(self):
        G=self.K3
        self.assertEquals(sorted(G.edges()),[(0,1),(0,2),(1,0),(1,2),(2,0),(2,1)])
        self.assertEquals(sorted(G.edges(0)),[(0,1),(0,2)])
        self.assertRaises((KeyError,GraphError), G.neighbors,-1)

    def test_edges_iter(self):
        G=self.K3
        self.assertEquals(sorted(G.edges()),[(0,1),(0,2),(1,0),(1,2),(2,0),(2,1)])
        self.assertEquals(sorted(G.edges_iter(0)),[(0,1),(0,2)])
        self.assertRaises((KeyError,GraphError), G.neighbors_iter,-1)


    def test_degree(self):
        G=self.K3
        self.assertEquals(G.degree(),[4,4,4])
        self.assertEquals(G.degree(with_labels=True),{0:4,1:4,2:4})
        self.assertEquals(G.degree(0),[4])
        self.assertEquals(G.degree(0,with_labels=True),{0:4})
        self.assertEquals(G.degree(-1),[])

    def test_degree_iter(self):
        G=self.K3
        self.assertEquals(list(G.degree_iter()),[(0,4),(1,4),(2,4)])
        self.assertEquals(dict(G.degree_iter()),{0:4,1:4,2:4})
        self.assertEquals(list(G.degree_iter(0)),[(0,4)])


    def test_in_degree(self):
        G=self.K3
        self.assertEquals(G.in_degree(),[2,2,2])
        self.assertEquals(G.in_degree(with_labels=True),{0:2,1:2,2:2})
        self.assertEquals(G.in_degree(0),2)
        self.assertEquals(G.in_degree(0,with_labels=True),{0:2})
        self.assertEquals(G.in_degree(-1),[])

    def test_in_degree_iter(self):
        G=self.K3
        self.assertEquals(list(G.in_degree_iter()),[(0,2),(1,2),(2,2)])
        self.assertEquals(dict(G.in_degree_iter()),{0:2,1:2,2:2})
        self.assertEquals(list(G.in_degree_iter(0)),[(0,2)])

    def test_out_degree(self):
        G=self.K3
        self.assertEquals(G.out_degree(),[2,2,2])
        self.assertEquals(G.out_degree(with_labels=True),{0:2,1:2,2:2})
        self.assertEquals(G.out_degree(0),2)
        self.assertEquals(G.out_degree(0,with_labels=True),{0:2})
        self.assertEquals(G.out_degree(-1),[])

    def test_out_degree_iter(self):
        G=self.K3
        self.assertEquals(list(G.out_degree_iter()),[(0,2),(1,2),(2,2)])
        self.assertEquals(dict(G.out_degree_iter()),{0:2,1:2,2:2})
        self.assertEquals(list(G.out_degree_iter(0)),[(0,2)])


    def test_size(self):
        G=self.K3
        self.assertEquals(G.size(),6)
        self.assertEquals(G.number_of_edges(),6)

def test_suite():
    return unittest.TestSuite((
        unittest.makeSuite(GraphTestCase),
        unittest.makeSuite(DiGraphTestCase),
        ))


def run_suite(verbosity=1):
    runner = unittest.TextTestRunner(verbosity=verbosity)
    runner.run(test_suite())

if __name__ == '__main__':
    run_suite()
