import unittest

from pylispng import lisp
from pylispng import sexpr
from pylispng import parser
from pylispng import symbol
from pylispng.util import DynamicList

class LispTestCase(unittest.TestCase):
    """

    """
    def setUp(self):
        """

        """
        self.exprs = [
            '(+ 5 3)',
            '(- 5 3)',
            '(- 3 5)',
            '(+ 5 2.3 2.7)',
            '(+ 5 (* 2 6.55))',
            '(* 5 (+ 3 4 6) 22 (* 2.4 5 0.001) (- 5 (+ 2 (- 2 4))))',
            "'(+ 5 3)",
            "'a",
            """'(this is "cool")""",
            "(first '(a b c))",
            "(rest '(a b c))",
            "(cons 'a '(b c d))",
            "(append '(a b c) '(d e f) (list 1 2 3 4))",
            "(setq fred '(a b c))",
            "(rest fred)",
            "(put 'fred 'birthday 'june-12-1973)",
            "(get 'fred 'birthday)",
            "(logic 0.5)",
            "(logic 1.0)",
            "(not (logic 0.0))",
            "(not (logic 0.35))",
            "(and *true* (not *true*))",
            "(or *true* (not *true*))",
            "(setq *true* (logic .95))",
            "(and *true* (not *true*))",
            "(or *true* (not *true*))",
            "(setq *true* (logic 1.0))",
            "(> 5 2)", 
            "(> 2 5)",
            "(> 5 (* 3 (+ 2 1)))",
            "(> (* 3 (+ 2 1)) (+ 1 2 3))",
            '(== "wow" "fred")',
            '(!= "wow" "fred")',
            "(>= 5 5)",
            '(<= "andy" "bobbie")',
            "(if (> 5 2) 'true 'false)",
            "(if (< 5 2) 'true 'false)",
            "(if (< 5 2) 'true)",
            "(if (not (< 5 2)) 'true)",
            "(lambda (x y) (* x y))",
            "((lambda (x y) (* x y)) 5 2)",
            "(lambda (x y) (* x y) (+ x y))",
            "((lambda (x y) (* x y) (+ x y)) 5 2)",
            "(def mult (lambda (x y) (* x y)))",
            "(mult 5 7)",
            "(def dumb (lambda (x y) (if (> x y) (* x y) (/ x y))))",
            "(dumb 5.0 3.0)",
            "(dumb 3.0 5.0)",
            "(let ((x 5) (y 3)) (+ x y))",
            "(let ((x 5) (y 3) z) (+ x y))",
            "(def dumb (lambda (x) (let ((y 12)) (* x y))))",
            "(dumb 3)",
            "`a",
            "`(a b ,fred)",
            "`(a b ,@fred)",
            "`(a (b (c (d (e (f ,@fred))))))",
            "((macro (x) `(+ ,x 5)) 4)",
            "(def incf (macro (x) `(setq ,x (+ ,x 1))))",
            "(setq whee 33)",
            "(incf whee)",
            "(def foo (lambda (x y) (if (<= y 0) x (begin (foo (cons y x) (- y 1))))))",
            "(foo '() 12)",
            "(let ((x 5)) (def bar (lambda () x)))",
            "(bar)",
            "(env-set (get-environment bar) 'x 33)",
            "(bar)",
            "(setq wow (list (cons 'a 'b) (cons 'q 'fun)))",
            "(assoc 'a wow)",
            '(- 5 (* 2 4))',
            '(- (* 2 4) 5)',
            '(/ 17 350)',
            '(/ 350 17)',
            ]
        self.answers = [
            '8',
            '2',
            '-2',
            '10.0',
            '18.1',
            '85.8',
            '(+ 5 3)',
            'a',
            "(this is 'cool')",
            'a',
            '(b c)',
            '(a b c d)',
            '(a b c d e f 1 2 3 4)',
            '(a b c)',
            '(b c)',
            'june-12-1973',
            'june-12-1973',
            '(logic 0.5)',
            '*true*',
            '*true*',
            '(logic 0.65)',
            '*false*',
            '*true*',
            '(logic 0.95)',
            '(logic 0.05)',
            '(logic 0.95)',
            '*true*',
            '*true*',
            '*false*',
            '*false*',
            '*true*',
            '*false*',
            '*true*',
            '*true*',
            '*true*',
            'true',
            'false',
            '*false*',
            'true',
            '(lambda (x y) (* x y))',
            '10',
            '(lambda (x y) (begin (* x y) (+ x y)))',
            '7',
            '(lambda (x y) (* x y))',
            '35',
            '(lambda (x y) (if (> x y) (* x y) (/ x y)))',
            '15.0',
            '0.6',
            '8',
            '8',
            '(lambda (x) (let ((y 12)) (* x y)))',
            '36',
            'a',
            '(a b (a b c))',
            '(a b a b c)',
            '(a (b (c (d (e (f a b c))))))',
            '9',
            '(macro (x) (i-quote (setq , x (+ , x 1))))',
            '33',
            '34',
            '(lambda (x y) (if (<= y 0) x (begin (foo (cons y x) (- y 1)))))',
            '(1 2 3 4 5 6 7 8 9 10 11 12)',
            '(lambda () x)',
            '5',
            '33',
            '33',
            '((a . b) (q . fun))',
            'b',
            '-3',
            '3',
            '0.0485714285714',
            '20.5882352941',
        ]
        self.reader = parser.Parser()
        self.lisper = lisp.Interpreter()

    def test_dataStructure(self):
        """

        """
        self.assertEquals(lisp.getType('+'), symbol.FunctionObject)
        self.assertEquals(lisp.getType('*'), symbol.FunctionObject)
        self.assertEquals(lisp.getType('and'), symbol.FunctionObject)
        self.assertEquals(lisp.getType('lambda'), symbol.SyntaxObject)
        self.assertEquals(lisp.getType('if'), symbol.SyntaxObject)
        self.assertEquals(lisp.getType('def'), symbol.SyntaxObject)
        self.assertEquals(lisp.getType('x'), symbol.SymbolObject)
        self.assertEquals(lisp.getType(5), symbol.SymbolObject)

    def test_parser(self):
        """
        As long as getSExpr() runs without throwing an exception, we're good.
        """
        for expr in self.exprs:
            sexpr1 = self.reader.getSExpr(expr)
            sexpr2 = parser.Parser(expr).sexpr
            self.assertTrue(isinstance(sexpr1, symbol.ListObject))
            self.assertEquals(str(sexpr1), str(sexpr2))

    def test_lisper(self):
        """

        """
        count = 0
        for expr, expected in zip(self.exprs, self.answers):
            count += 1
            sexpr = self.reader.getSExpr(expr)
            answer = self.lisper.eval(sexpr)
            self.assertEquals(str(answer), expected)

    def test_expressionLengths(self):
        """
        PyLisp determines expression length by the number of immediate
        children, not all children recursively.
        """
        answers = [3, 3, 3, 4, 3, 6]
        for expr, expected in zip(self.exprs[:6], answers):
            sexpr = self.reader.getSExpr(expr)
            self.assertEquals(len(sexpr), expected)

    def test_sExpressions(self):
        """

        """
        l = sexpr.SExpression()
        l.append('+')
        l.append('3')
        l.append('5')
        l.append('(* 3 3)')
        form = '(+ 3 5 (* 3 3))'
        value = 17
        depth = 2
        length = 4
        size = 2
        self.assertEquals(str(l), form)
        self.assertEquals(l.eval(), value)
        self.assertEquals(l.getDepth(), depth)
        self.assertEquals(len(l), length)
        self.assertEquals(l.getSize(), size)
        # more complicated expressions, chunk at a time
        l = sexpr.SExpression('(+ 4 19)')
        l.append('(- 2 (* 45 3))')
        l.append('(* 7 (- 37 99))')
        form = '(+ 4 19 (- 2 (* 45 3)) (* 7 (- 37 99)))'
        value = -544
        depth = 3
        length = 5
        size = 5
        self.assertEquals(str(l), form)
        self.assertEquals(l.eval(), value)
        self.assertEquals(l.getDepth(), depth)
        self.assertEquals(len(l), length)
        self.assertEquals(l.getSize(), size)

    def test_sExpressionsWithInts(self):
        # now test with integers
        l = sexpr.SExpression()
        l.append('+')
        l.append(3)
        l.append(5)
        l.append('(* 3 3)')
        form = '(+ 3 5 (* 3 3))'
        value = 17
        self.assertEquals(str(l), form)
        self.assertEquals(l.eval(), value)
